import logging
import re
import zlib
from struct import pack, unpack
from typing import Callable, Iterator, Optional, Tuple

from _hashlib import HASH as Hashobj

from ..exceptions import ParseError
from ..file import BufferedBinaryIoT, read_or_raise

png_sig = b"\x89PNG\r\n\x1a\n"
chunk_type_p = re.compile(rb"^[a-zA-Z]{4}$")

image_chunks = {b"IHDR", b"IDAT", b"PLTE", b"tRNS", b"acTL", b"fcTL", b"fdAT", b"IEND"}
image_meta_chunks = {b"sBIT", b"pHYs", b"vpAg", b"gAMA", b"sRGB", b"cHRM", b"iCCP"}
meta_data_chunks = {b"iTXt", b"tEXt", b"zTXt", b"tIME", b"bKGD", b"eXIf", b"dSIG"}


def png_chunk(chunk_type_ascii: str, chunk: bytes) -> bytes:
    chunk_type = chunk_type_ascii.encode("ascii")

    length = pack("!I", len(chunk))
    _crc = zlib.crc32(chunk_type)
    _crc = zlib.crc32(chunk, _crc)
    crc = pack("!i", _crc)
    return length + chunk_type + chunk + crc


def IHDR(width: int, height: int) -> bytes:
    bitdepth = 1
    colortype = 0  # grayscale
    compression = 0
    filter = 0
    interlace = 0
    chunk = pack("!IIBBBBB", width, height, bitdepth, colortype, compression, filter, interlace)

    return png_chunk("IHDR", chunk)


def IEND() -> bytes:
    return png_chunk("IEND", b"")


def IDAT(binary: bytes, level: int = 9) -> bytes:
    # ignores filter for scanlines
    return png_chunk("IDAT", zlib.compress(binary, level))


def binary2png(binary: bytes, width: int, height: int) -> bytes:
    return png_sig + IHDR(width, height) + IDAT(binary) + IEND()


def iter_png_fp(stream: BufferedBinaryIoT, translate: bool = True, verify_crc: bool = True) -> Iterator[tuple]:
    """Parses PNG / APNG files."""

    signature = read_or_raise(stream, 8)
    if signature != png_sig:
        raise ParseError("Not a png file")

    if not translate:
        yield (b"", b"", signature, b"")

    iend = False

    while True:
        data = stream.read(8)

        if not data:
            if iend:
                return
            else:
                raise EOFError
        elif len(data) != 8:
            raise EOFError

        length, chunk_type = unpack(">I4s", data)
        if not chunk_type_p.match(chunk_type):
            raise ParseError(f"Invalid chunk type: {chunk_type} at {stream.tell() - 4}")

        chunk_type_ascii = chunk_type.decode("ascii")

        if chunk_type_ascii == "IEND":
            iend = True

        if length > 0:
            chunk = read_or_raise(stream, length)
        else:
            chunk = b""

        data_crc = read_or_raise(stream, 4)
        (crc,) = unpack(">I", data_crc)

        if verify_crc:
            _crc = zlib.crc32(chunk_type)
            _crc = zlib.crc32(chunk, _crc)
            if crc != _crc:
                raise ParseError("Invalid CRC")

        if translate:
            yield chunk_type_ascii, chunk, crc
        else:
            yield data[0:4], data[4:8], chunk, data_crc


def iter_png(path: str, translate: bool = True, verify_crc: bool = True) -> Iterator[tuple]:
    """Same as `iter_png_fp()` except that it accepts a path."""

    with open(path, "rb") as fr:
        yield from iter_png_fp(fr, translate=translate, verify_crc=verify_crc)


def copy_png_fp(
    fin: BufferedBinaryIoT,
    fout: BufferedBinaryIoT,
    filter_chunks: Optional[Callable[[bytes], bool]] = None,
    verify_crc: bool = False,
) -> None:
    """Same as `copy_png()` except that it accepts file-like objects."""

    filter_chunks = filter_chunks or (lambda chunk_type: True)

    for length, chunk_type, chunk, crc in iter_png_fp(fin, translate=False, verify_crc=verify_crc):
        if filter_chunks(chunk_type):
            fout.write(length)
            fout.write(chunk_type)
            fout.write(chunk)
            fout.write(crc)


def copy_png(
    inpath: str, outpath: str, filter_chunks: Optional[Callable[[bytes], bool]] = None, verify_crc: bool = False
) -> None:
    """Copy PNG file `inpath` to `outpath` while ignoring the PNG chunks
    given by `filter_chunks`.

    Example:
            copy_jpeg("in.png", "out.png", lambda chunk: chunk not in {"tEXt", "tIME"})
    """

    with open(inpath, "rb") as fr, open(outpath, "xb") as fw:
        copy_png_fp(fr, fw, filter_chunks=filter_chunks, verify_crc=verify_crc)


def hash_raw_png(path: str, hashobj: Hashobj) -> None:
    """Create a hash of the JPEG at `path` skipping over meta data sections."""

    def filter_chunks(ct: bytes) -> bool:
        return ct in image_chunks

    for length, chunk_type, chunk, _crc in iter_png(path, translate=False, verify_crc=False):
        if filter_chunks(chunk_type):
            hashobj.update(length)
            hashobj.update(chunk_type)
            hashobj.update(chunk)


def parse_tEXt(chunk: bytes) -> Tuple[str, str]:
    keyword, text = chunk.split(b"\0")
    return keyword.decode("latin1"), text.decode("latin1")


def parse_tIME(chunk: bytes) -> Tuple[int, int, int, int, int, int]:
    year, month, day, hour, minute, second = unpack("!HBBBBB", chunk)

    return year, month, day, hour, minute, second


if __name__ == "__main__":
    from argparse import ArgumentParser

    from genutility.args import is_dir
    from genutility.iter import consume

    parser = ArgumentParser()
    parser.add_argument("path", type=is_dir)
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-r", "--recursive", action="store_true")
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    if args.recursive:
        it = args.path.rglob("*.png")
    else:
        it = args.path.glob("*.png")

    valid = 0
    invalid = 0

    for path in it:
        try:
            consume(iter_png(path))

        except ParseError as e:
            logging.debug("ParseError in %s: %s", path, e)
            invalid += 1

        except EOFError:
            logging.debug("EOFError in %s", path)
            invalid += 1

        else:
            valid += 1

    print("valid", valid, "invalid", invalid)



from ctypes import byref, create_unicode_buffer, sizeof
from ctypes.wintypes import DWORD, LARGE_INTEGER
from enum import IntFlag
from msvcrt import get_osfhandle, open_osfhandle
from os import fdopen
from typing import Optional

from cwinsdk import struct2dict
from cwinsdk.km import ntddk, ntifs, wdm
from cwinsdk.shared.minwindef import MAX_PATH
from cwinsdk.um import fileapi, winnt
from cwinsdk.um.handleapi import CloseHandle
from cwinsdk.um.winnt import GENERIC_READ, GENERIC_WRITE
from typing_extensions import Self

_mode2access = {"": 0, "r": GENERIC_READ, "w": GENERIC_WRITE, "r+": GENERIC_READ | GENERIC_WRITE}


class FileSystem(IntFlag):
    CASE_SENSITIVE_SEARCH = winnt.FILE_CASE_SENSITIVE_SEARCH
    CASE_PRESERVED_NAMES = winnt.FILE_CASE_PRESERVED_NAMES
    UNICODE_ON_DISK = winnt.FILE_UNICODE_ON_DISK
    PERSISTENT_ACLS = winnt.FILE_PERSISTENT_ACLS
    FILE_COMPRESSION = winnt.FILE_FILE_COMPRESSION
    VOLUME_QUOTAS = winnt.FILE_VOLUME_QUOTAS
    SUPPORTS_SPARSE_FILES = winnt.FILE_SUPPORTS_SPARSE_FILES
    SUPPORTS_REPARSE_POINTS = winnt.FILE_SUPPORTS_REPARSE_POINTS
    SUPPORTS_REMOTE_STORAGE = winnt.FILE_SUPPORTS_REMOTE_STORAGE
    RETURNS_CLEANUP_RESULT_INFO = winnt.FILE_RETURNS_CLEANUP_RESULT_INFO
    SUPPORTS_POSIX_UNLINK_RENAME = winnt.FILE_SUPPORTS_POSIX_UNLINK_RENAME
    VOLUME_IS_COMPRESSED = winnt.FILE_VOLUME_IS_COMPRESSED
    SUPPORTS_OBJECT_IDS = winnt.FILE_SUPPORTS_OBJECT_IDS
    SUPPORTS_ENCRYPTION = winnt.FILE_SUPPORTS_ENCRYPTION
    NAMED_STREAMS = winnt.FILE_NAMED_STREAMS
    READ_ONLY_VOLUME = winnt.FILE_READ_ONLY_VOLUME
    SEQUENTIAL_WRITE_ONCE = winnt.FILE_SEQUENTIAL_WRITE_ONCE
    SUPPORTS_TRANSACTIONS = winnt.FILE_SUPPORTS_TRANSACTIONS
    SUPPORTS_HARD_LINKS = winnt.FILE_SUPPORTS_HARD_LINKS
    SUPPORTS_EXTENDED_ATTRIBUTES = winnt.FILE_SUPPORTS_EXTENDED_ATTRIBUTES
    SUPPORTS_OPEN_BY_FILE_ID = winnt.FILE_SUPPORTS_OPEN_BY_FILE_ID
    SUPPORTS_USN_JOURNAL = winnt.FILE_SUPPORTS_USN_JOURNAL
    SUPPORTS_INTEGRITY_STREAMS = winnt.FILE_SUPPORTS_INTEGRITY_STREAMS
    SUPPORTS_BLOCK_REFCOUNTING = winnt.FILE_SUPPORTS_BLOCK_REFCOUNTING
    SUPPORTS_SPARSE_VDL = winnt.FILE_SUPPORTS_SPARSE_VDL
    DAX_VOLUME = winnt.FILE_DAX_VOLUME
    SUPPORTS_GHOSTING = winnt.FILE_SUPPORTS_GHOSTING


class WindowsHandle:
    def __init__(self, handle: int, doclose: bool = True, timeout_ms: Optional[int] = None) -> None:
        if not isinstance(handle, int):
            raise TypeError("handle must be an int")

        self.handle = handle
        self.doclose = doclose
        self.timeout_ms = timeout_ms

    @property
    def overlapped(self) -> True:
        return self.timeout_ms is not None

    @classmethod
    def from_file(cls, fp):
        return cls.from_fd(fp.fileno())

    @classmethod
    def from_fd(cls, fd):
        return cls(get_osfhandle(fd), doclose=False)

    def get_fd(self, flags):
        return open_osfhandle(self.handle, flags)

    def get_file(self, flags, mode="r", buffering=-1, encoding=None, errors=None, newline=None, closefd=True):
        return fdopen(self.get_fd(flags), mode, buffering, encoding, errors, newline, closefd)

    def close(self) -> None:
        CloseHandle(self.handle)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.doclose:
            self.close()

    def filesize(self) -> int:
        FileSize = LARGE_INTEGER()
        fileapi.GetFileSizeEx(self.handle, byref(FileSize))
        return FileSize

    def file_information(self) -> dict:
        FileInformation = fileapi.BY_HANDLE_FILE_INFORMATION()
        fileapi.GetFileInformationByHandle(self.handle, byref(FileInformation))
        return struct2dict(FileInformation)

    def volume_information(self) -> dict:
        VolumeNameBuffer = create_unicode_buffer(MAX_PATH + 1)
        VolumeNameSize = MAX_PATH + 1
        VolumeSerialNumber = DWORD()
        MaximumComponentLength = DWORD()
        FileSystemFlags = DWORD()
        FileSystemNameBuffer = create_unicode_buffer(MAX_PATH + 1)
        FileSystemNameSize = MAX_PATH + 1

        fileapi.GetVolumeInformationByHandleW(
            self.handle,
            VolumeNameBuffer,
            VolumeNameSize,
            byref(VolumeSerialNumber),
            byref(MaximumComponentLength),
            byref(FileSystemFlags),
            FileSystemNameBuffer,
            FileSystemNameSize,
        )

        return {
            "VolumeName": VolumeNameBuffer.value,
            "VolumeSerialNumber": VolumeSerialNumber.value,
            "MaximumComponentLength": MaximumComponentLength.value,
            "FileSystemFlags": FileSystem(FileSystemFlags.value),
            "FileSystemName": FileSystemNameBuffer.value,
        }

    def fs_full_size(self) -> dict:
        IoStatusBlock = wdm.IO_STATUS_BLOCK()
        FsInformation = ntddk.FILE_FS_FULL_SIZE_INFORMATION()

        ntifs.NtQueryVolumeInformationFile(
            self.handle,
            byref(IoStatusBlock),
            byref(FsInformation),
            sizeof(ntddk.FILE_FS_FULL_SIZE_INFORMATION),
            wdm.FS_INFORMATION_CLASS.FileFsFullSizeInformation,
        )
        return struct2dict(FsInformation)



from itertools import islice
from typing import Any, Callable, Collection, Iterable, Iterator, NamedTuple, Optional, TextIO, Tuple, Union

import numpy as np
from gensim.models.keyedvectors import KeyedVectors as KeyedVectorsOriginal

from .file import PathOrTextIO


class Vocab(NamedTuple):
    index: int
    count: int


class DuplicateEntry(ValueError):
    pass


def false(x: Any) -> bool:
    return False


class KeyedVectors(KeyedVectorsOriginal):
    """Enhancement of the original `gensim` `KeyedVectors` class. Supports loading of `glove` and
    `muse` word vector files.
    """

    @classmethod
    def load_a_format(
        cls,
        fin: Iterable[str],
        vocab_size: int,
        vector_size: int,
        datatype: Any = np.float32,
        discard: Optional[Callable[[str], bool]] = None,
    ) -> KeyedVectorsOriginal:
        discard = discard or false

        result = cls(vector_size)
        result.vector_size = vector_size
        result.vectors = np.zeros((vocab_size, vector_size), dtype=datatype)
        result.expandos["count"] = np.zeros(vocab_size, dtype=np.int64)

        for line in fin:
            word, vect = line.rstrip().split(" ", 1)

            if discard(word):
                continue

            weights = np.fromstring(vect, sep=" ", dtype=np.float32)
            # raise ValueError("invalid vector on line %s; is vector_size incorrect or file otherwise damaged?" % (i+1,))

            if word in result.key_to_index:
                raise DuplicateEntry(word)

            word_id = len(result.index_to_key)
            result.key_to_index[word] = word_id
            result.expandos["count"][word_id] = vocab_size - word_id
            result.vectors[word_id] = weights
            result.index_to_key.append(word)

        real_size = len(result.index_to_key)

        if real_size != vocab_size:
            if discard is None:
                raise EOFError("unexpected end of input; is vocab_size incorrect or file otherwise damaged?")
            else:
                result.vectors.resize((real_size, vector_size))  # this should be no-copy
                result.expandos["count"].resize(real_size)  # this should be no-copy

        return result

    def add(self, wwc: Collection[Tuple[str, np.ndarray, int]]) -> Iterator[Vocab]:
        """`wwc` is a collection of (word, weights, count) tuples"""

        vocab_size, vector_size = self.vectors.shape
        self.vectors.resize((vocab_size + len(wwc), vector_size))
        self.expandos["count"].resize(vocab_size + len(wwc))

        for i, (word, weights, count) in enumerate(wwc):
            word_id = vocab_size + i
            self.key_to_index[word] = word_id
            self.expandos["count"][word_id] = count
            self.vectors[word_id] = weights
            self.index_to_key.append(word)
            yield Vocab(word_id, count)

    def add_word(self, word: str, weights: Optional[np.ndarray] = None, count: int = 1):
        if weights is None:
            vocab_size, vector_size = self.vectors.shape
            weights = np.random.randn(1, vector_size)

        return list(self.add([(word, weights, count)]))[0]

    @classmethod
    def load_muse_format(
        cls, fname: Union[str, TextIO], limit: Optional[int] = None, discard: Optional[Callable[[str], bool]] = None
    ) -> KeyedVectorsOriginal:
        """reads at most `limit` vectors from muse file, if `discard` is not None it might be less.
        discard is a callable which should return False if word should not be loaded from file.
        """

        with PathOrTextIO(fname, "rt", encoding="utf-8", newline="\n") as fin:
            vocab_size_, vector_size = map(int, next(fin).split(" "))
            if limit is None:
                vs = vocab_size_
            else:
                vs = min(vocab_size_, limit)

            try:
                return cls.load_a_format(islice(fin, limit), vs, vector_size, discard=discard)
            except DuplicateEntry as e:
                raise ValueError(f"duplicate word '{e}' in {fname}")

    @classmethod
    def load_glove_format(
        cls, fname: Union[str, TextIO], limit: int, vector_size: int, discard: Optional[Callable[[str], bool]] = None
    ) -> KeyedVectorsOriginal:
        """Reads at most `limit` vectors from glove file, if `discard` is not None it might be less.
        `discard` is a callable which should return False if word should not be loaded from file.
        """

        with PathOrTextIO(fname, "rt", encoding="utf-8") as fin:
            try:
                return cls.load_a_format(islice(fin, limit), limit, vector_size, discard=discard)
            except DuplicateEntry as e:
                raise ValueError(f"duplicate word '{e}' in {fname}")

    def get_keras_embedding(self, train_embeddings: bool = False, mask_zero: bool = True) -> "Embedding":  # noqa: F821
        try:
            from keras.layers import Embedding
        except ImportError:
            raise ImportError("Please install Keras to use this function")

        if mask_zero:
            self.index_to_key = [None] + self.index_to_key
            zero_vec = np.zeros((1, self.vector_size))
            self.vectors = np.concatenate([zero_vec, self.vectors], axis=0)

        weights = self.vectors

        # set `trainable` as `False` to use the pretrained word embedding
        # No extra mem usage here as `Embedding` layer doesn't create any new matrix for weights
        layer = Embedding(
            input_dim=weights.shape[0],
            output_dim=weights.shape[1],
            weights=[weights],
            trainable=train_embeddings,
            mask_zero=mask_zero,
        )
        return layer

    def transform_words_to_indices(self, words: Collection[str]) -> Iterator[int]:
        return (self.key_to_index[word] for word in words)

    def transform_words_to_indices_tuple(self, words: Collection[str]) -> Tuple[int, ...]:
        return tuple(self.transform_words_to_indices(words))


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--glove-file")
    parser.add_argument("--muse-file")
    parser.add_argument("--dimensions", type=int, default=300)
    args = parser.parse_args()

    if args.glove_file:
        word_vectors = KeyedVectors.load_glove_format(args.glove_file, 400000, args.dimensions)
        print(word_vectors.most_similar(positive=["woman", "king"], negative=["man"]))

    if args.muse_file:
        word_vectors = KeyedVectors.load_muse_format(args.muse_file)
        print(word_vectors.most_similar(positive=["woman", "king"], negative=["man"]))



import sys
from numbers import Number
from time import time
from typing import Iterable, Optional, Sequence, TextIO, Union

from .iter import _lstr, progressdata


class progress_content:
    def __init__(
        self,
        it: Union[Iterable, Sequence],
        length: Optional[int] = None,
        refresh: Number = 1,
        file: Optional[TextIO] = sys.stdout,
    ) -> None:
        self.it = it
        self.length = length
        self.refresh = refresh
        self.file = file

    def __iter__(self):
        return progressdata(self.it, self.length, self.refresh, file=self.file)

    def __aiter__(self):
        return self.AsyncIterProgress(self.it, self.length, self.refresh, file=self.file)

    class AsyncIterProgress:
        def __init__(
            self, it: Union[Iterable, Sequence], length: Optional[int], refresh: Number, file: Optional[TextIO]
        ):
            self.it = it.__aiter__()
            self.refresh = refresh
            self.file = file

            self.length, self.lstr = _lstr(self.it, length)
            self.last = self.start = time()
            self.total = 0

        async def __anext__(self):
            try:
                elm = await self.it.__anext__()
                self.total += len(elm)
                current = time()
                if current - self.last > self.refresh:
                    self.last = current
                    duration = current - self.start
                    print(
                        f"{self.total}{self.lstr}, running for {int(duration)} seconds ({self.total/duration:0.2e}/s).",
                        end="\r",
                        file=self.file,
                    )
                return elm

            except StopAsyncIteration:
                print(f"Finished {self.total} in {int(self.last - self.start)} seconds.", end="\r", file=self.file)
                raise StopAsyncIteration


if __name__ == "__main__":
    import asyncio

    class gensync:
        def __init__(self):
            self.i = 3

        def __iter__(self):
            return self

        def __next__(self):
            if self.i > 0:
                self.i -= 1
                return "asd"
            else:
                raise StopIteration

    class genasync:
        def __init__(self):
            self.i = 3

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.i > 0:
                self.i -= 1
                return "asd"
            else:
                raise StopAsyncIteration

    for _i in progress_content(gensync()):
        pass

    print()

    async def main():
        async for _i in progress_content(genasync()):
            pass

    asyncio.run(main())



from collections import defaultdict
from contextlib import nullcontext
from difflib import unified_diff
from functools import wraps
from itertools import chain, product, zip_longest
from typing import IO, Any, Callable, Collection, DefaultDict, Hashable, Iterable, Mapping, Optional, Tuple, TypeVar
from unittest import TestCase
from unittest.result import TestResult
from unittest.runner import TextTestRunner

from .file import equal_files
from .iter import progress

T = TypeVar("T")
U = TypeVar("U")


anullcontext = nullcontext()


def print_file_diff(fromfile: str, tofile: str, encoding: str) -> None:
    with open(fromfile, encoding=encoding) as fr:
        fromlines = fr.readlines()
    with open(tofile, encoding=encoding) as fr:
        tolines = fr.readlines()

    for line in unified_diff(fromlines, tolines, fromfile, tofile):
        print(line, end="")


class MyTestResult(TestResult):
    """Extension of TestResult to support numbering test cases"""

    def __init__(self, stream: IO[str], descriptions, verbosity):
        """Initializes the test number generator, then calls super impl"""
        self.stream = stream
        self.current_test = 0
        self.current_subtest = 0
        TestResult.__init__(self, stream, descriptions, verbosity)

    def progress(self) -> None:
        self.stream.write(f"Running test #{self.current_test}, subtest #{self.current_subtest}\r")

    def startTest(self, test) -> None:
        self.current_test += 1
        TestResult.startTest(self, test)
        self.progress()

    def stopTest(self, test) -> None:
        TestResult.stopTest(self, test)

    def stopTestRun(self) -> None:
        self.stream.write("\n")
        TestResult.stopTestRun(self)

    def addSubTest(self, test, subtest, outcome) -> None:
        """Called when a subtest finishes."""

        TestResult.addSubTest(self, test, subtest, outcome)
        self.testsRun += 1
        self.current_subtest += 1
        self.progress()


class MyTestRunner(TextTestRunner):
    """Extension of TestRunner to support numbering test cases"""

    resultclass = MyTestResult

    def run(self, test):
        """Stores the total count of test cases, then calls super impl"""

        self.total_test_cases = test.countTestCases()
        return TextTestRunner.run(self, test)

    def _makeResult(self):
        return TextTestRunner._makeResult(self)


class NoRaise:
    def __init__(self, testcase: TestCase, message: Optional[str] = None) -> None:
        self.testcase = testcase
        self.message = message

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            if self.message:
                self.testcase.fail(self.message)
            else:
                self.testcase.fail(exc_value)


def make_comparable(d: Any) -> Any:
    if isinstance(d, (str, int, type(None))):
        return d
    elif isinstance(d, list):
        return list(make_comparable(i) for i in d)
    elif isinstance(d, tuple):
        return tuple(make_comparable(i) for i in d)
    elif isinstance(d, dict):
        return make_comparable(list(d.items()))
    else:
        raise TypeError(f"must be list, tuple or dict, not {type(d)}")


def is_equal_unordered(seq_a: Collection, seq_b: Collection) -> bool:
    if isinstance(seq_a, set) and isinstance(seq_b, set):
        return seq_a == seq_b

    seq_a = sorted(make_comparable(seq_a))
    seq_b = sorted(make_comparable(seq_b))
    return seq_a == seq_b


class MyTestCase(TestCase):
    def assertAnd(self, first: Any, second: Any, msg: Optional[str] = None) -> None:
        self.assertTrue(first and second, msg)

    def assertTypeEqual(self, first: Any, second: Any, msg: Optional[str] = None) -> None:
        self.assertEqual(type(first), type(second), msg)

    def assertFilesEqual(self, first_path: Any, second_path: Any, msg: Optional[str] = None) -> None:
        self.assertTrue(equal_files(first_path, second_path), msg)

    def assertIterEqual(self, first: Iterable, second: Iterable, msg: Optional[str] = None) -> None:
        for i, (a, b) in enumerate(zip_longest(first, second)):
            if msg:
                msg = " : " + str(msg)
            self.assertEqual(a, b, msg=f"in iteration index {i}: {msg}")

    def assertIterAlmostEqual(self, first: Iterable, second: Iterable, msg: Optional[str] = None) -> None:
        for i, (a, b) in enumerate(zip_longest(first, second)):
            if msg:
                msg = " : " + str(msg)
            try:
                self.assertAlmostEqual(a, b, msg=f"in iteration index {i}: {msg}")
            except TypeError:
                raise AssertionError("Invalid types (probably different length iters)")  # from None

    def assertAllEqual(self, args: Iterable, msg: Optional[str] = None) -> None:  # *args doesn't work in python2!?
        it = iter(args)
        first = next(it)
        for second in it:
            self.assertEqual(first, second, msg)

    def assertIterIterEqual(
        self, first: Iterable[Iterable], second: Iterable[Iterable], msg: Optional[str] = None
    ) -> None:  # UNTESTED
        for i, (a, b) in enumerate(zip_longest(first, second)):
            if msg:
                msg = " : " + str(msg)
            self.assertIterEqual(a, b, msg=f"in iteration index {i}: {msg}")

    def assertPriorityEqual(
        self,
        first: Iterable[Optional[Tuple[Hashable, Hashable]]],
        second: Iterable[Optional[Tuple[Hashable, Hashable]]],
        msg: Optional[str] = None,
    ) -> None:
        d1: DefaultDict[Hashable, set] = defaultdict(set)
        d2: DefaultDict[Hashable, set] = defaultdict(set)

        for a, b in zip_longest(first, second):
            if a is not None:
                val_1, p_1 = a
                if b is None and p_1 == 0:
                    continue

            if b is not None:
                val_2, p_2 = b
                if a is None and p_2 == 0:
                    continue

            self.assertAnd(a, b)

            if val_1 == 0 and val_2 == 0:
                continue

            d1[p_1].add(val_1)
            d2[p_2].add(val_2)
        self.assertEqual(d1, d2, msg)

    def assertUnorderedSeqEqual(self, first: Iterable, second: Iterable, msg: Optional[str] = None) -> None:
        # fixme: use is_equal_unordered

        first = sorted(first)
        second = sorted(second)
        self.assertEqual(first, second, msg)

    def assertNoRaise(self, msg: Optional[str] = None) -> NoRaise:
        return NoRaise(self, msg)

    def assertEqualsOneOf(self, result: T, truths: Iterable[T]) -> None:
        for truth in truths:
            if result == truth:
                return

        raise AssertionError("Result does not match one of the provided truths")  # from None

    def assertUnorderedMappingEqual(
        self, first: Mapping[T, U], second: Mapping[T, U], msg: Optional[str] = None
    ) -> None:
        self.assertEqual(len(first), len(second), msg)

        for a, b in zip(sorted(first.keys()), sorted(second.keys())):
            self.assertEqual(a, b)
            self.assertEqual(first[a], second[b])


def random_arguments(n: int, *funcs: Callable[[], Any]) -> Callable[[Callable], Callable]:
    def decorator(func):
        @wraps(func)
        def inner(self):
            for i in range(n):
                with self.subTest(str(i)):
                    if func(self, *(f() for f in funcs)) is not None:
                        raise AssertionError

        return inner

    return decorator


# also called: parameterize
def parametrize(*args_list: tuple) -> Callable[[Callable], Callable]:
    def decorator(func):
        @wraps(func)
        def inner(self):
            for args in args_list:
                with self.subTest(str(args)[:1000]):
                    if func(self, *args) is not None:
                        raise AssertionError

        return inner

    return decorator


def parametrize_product(*args_list: Iterable) -> Callable[[Callable], Callable]:
    def decorator(func):
        @wraps(func)
        def inner(self):
            for args in product(*args_list):
                with self.subTest(str(args)):
                    if func(self, *args) is not None:
                        raise AssertionError

        return inner

    return decorator


def parametrize_starproduct(*args_list: Iterable[Iterable]) -> Callable[[Callable], Callable]:
    def decorator(func):
        @wraps(func)
        def inner(self):
            for args in product(*args_list):
                combined_args = tuple(chain.from_iterable(args))
                with self.subTest(str(combined_args)):
                    if func(self, *combined_args) is not None:
                        raise AssertionError

        return inner

    return decorator


def repeat(number: int, verbose: bool = False) -> Callable[[Callable], Callable]:
    def decorator(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            for _i in progress(range(number), disable=not verbose):
                if func(self, *args, **kwargs) is not None:  # no self.subTest(str(i))
                    raise AssertionError

        return inner

    return decorator



from typing import Callable, MutableSequence, Optional, TypeVar

T = TypeVar("T")


def _insertion(seq: MutableSequence, cmp_: Callable, left: int, right: int, gap: int) -> None:
    loc = left + gap
    while loc <= right:
        i = loc - gap
        value = seq[loc]
        while i >= left and cmp_(seq[i], value) > 0:
            seq[i + gap] = seq[i]
            i -= gap
        seq[i + gap] = value
        loc += gap


GROUP_SIZE = 5


def cmp(x: T, y: T) -> int:
    """
    Return negative if x<y, zero if x==y, positive if x>y.
    """
    return (x > y) - (x < y)


def median_of_medians(
    seq: MutableSequence, cmp_: Optional[Callable] = None, left: int = 0, right: Optional[int] = None, depth: int = 0
) -> int:
    """Approximate median selection algorithm.
    This is not the full median of medians algorithm.
    Currently only works for len(seq) == GROUP_SIZE ** x
    """

    offset = (GROUP_SIZE ** (depth + 1) - GROUP_SIZE**depth) // 2

    gap = GROUP_SIZE**depth

    cmp_ = cmp_ or cmp

    if not right:
        right = len(seq) - 1

    span = GROUP_SIZE * gap
    num = (right - left + 1) // span

    if num == 0:
        _insertion(seq, cmp_, left, right, gap)
        num = (right - left + 1) // gap
        return left + gap * (num - 1) // 2  # select median

    s = left
    while s < right:
        _insertion(seq, cmp_, s, s + span - 1, gap)
        s += span

    if num < GROUP_SIZE:
        _insertion(seq, cmp_, left + offset, right, span)
        return left + offset + num * span // 2  # select median
    else:
        return median_of_medians(seq, cmp_, left + offset, right, depth + 1)



import logging
from collections import defaultdict
from functools import wraps
from typing import Any, Callable, Iterable, Optional


def printr(*objs: Any, end: str = "\n", depth: int = 0) -> None:
    for i, obj in enumerate(objs, 1):
        if isinstance(obj, str):
            print(obj, end="")
        elif isinstance(obj, dict):
            print("{", end="")
            for j, (k, v) in enumerate(obj.items(), 1):
                printr(k, end=end, depth=depth + 1)
                print(": ", end="")
                printr(v, end=end, depth=depth + 1)
                if j != len(obj):
                    print(", ", end="")
            print("}", end="")
        else:
            try:
                print("(", end="")
                for j, item in enumerate(obj, 1):
                    printr(item, end=end, depth=depth + 1)
                    if j != len(obj):
                        print(", ", end="")
                print(")", end="")
            except TypeError:
                print(obj, end="")

        if i != len(objs):
            print(" ", end="")

    if depth == 0:
        print(end=end)


def _to_union(types: Iterable[str]) -> str:
    types = list(set(types))

    if not types:
        return ""
    elif len(types) == 1:
        return types[0]
    elif len(types) > 1:
        return "Union[" + ", ".join(types) + "]"

    assert False


def _type_str(obj: Any) -> str:
    return type(obj).__name__


def rec_repr(obj: Any) -> str:
    if isinstance(obj, defaultdict):
        return _type_str(obj) + "[" + rec_repr(obj.default_factory) + "]"
    elif isinstance(obj, list):
        return "List[" + _to_union(map(rec_repr, obj)) + "]"
    elif isinstance(obj, set):
        return "Set[" + _to_union(map(rec_repr, obj)) + "]"
    else:
        return _type_str(obj)


def _arg_str(arg: Any, maxlen: Optional[int] = None, app: str = "...", repr_args: bool = True) -> str:
    if repr_args:
        arg = repr(arg)

    assert isinstance(arg, str)

    if maxlen:
        if len(arg) <= maxlen + len(app):
            return arg
        else:
            return arg[:maxlen] + app
    else:
        return arg


def _kwarg_str(key: str, value: Any, maxlen: Optional[int] = None, app: str = "...", repr_args: bool = True) -> str:
    return key + "=" + _arg_str(value, maxlen, app, repr_args)


def args_str(args: tuple, kwargs: dict, maxlen: Optional[int] = 20, app: str = "...", repr_args: bool = True) -> str:
    """Creates printable string from function arguments.
    If the string needs to be truncated to fit `maxlen`, `app` will be appended.
    The length of `app` is not included in `maxlen`.
    If the original string is shorter or equal `maxlen + len(app)`,
    it will be returned unmodified.
    """

    args = ", ".join(_arg_str(arg, maxlen, app, repr_args) for arg in args)
    kwargs = ", ".join(_kwarg_str(k, v, maxlen, app, repr_args) for k, v in kwargs.items())

    if args:
        if kwargs:
            return args + ", " + kwargs
        else:
            return args
    else:
        if kwargs:
            return kwargs
        else:
            return ""


def log_call(s: str) -> Callable:
    """Decorator to log function calls using template string `s`.
    Available format fields are: 'name', 'args' and 'kwargs'.
    """

    def dec(func: Callable) -> Callable:
        def inner(*args, **kwargs):
            logging.debug(s.format(name=func.__name__, args=args, kwargs=kwargs))
            return func(*args, **kwargs)

        return inner

    return dec


def log_wrap_call(func: Callable) -> Callable:
    """Decorator which logs all calls to `func` with all arguments."""

    @wraps(func)
    def inner(*args, **kwargs):
        logging.debug("START %s(%s)", func.__name__, args_str(args, kwargs))

        try:
            ret = func(*args, **kwargs)
        except BaseException:
            logging.exception("RAISED %s(%s)", func.__name__, args_str(args, kwargs))
            raise

        logging.debug("END %s(%s)", func.__name__, args_str(args, kwargs))
        return ret

    return inner


def log_methodcall(func: Callable) -> Callable:
    """Decorator to log method calls with arguments."""

    @wraps(func)
    def inner(self, *args, **kwargs):
        classname = self.__class__.__name__
        # classname = type(self).__name__ ?
        logging.debug("%s.%s(%s)", classname, func.__name__, args_str(args, kwargs))
        return func(self, *args, **kwargs)

    return inner


def log_methodcall_result(func: Callable) -> Callable:
    """Decorator to log method calls with arguments and results."""

    @wraps(func)
    def inner(self, *args, **kwargs):
        classname = self.__class__.__name__
        # classname = type(self).__name__ ?
        logging.debug("%s.%s(%s)", classname, func.__name__, args_str(args, kwargs))
        res = func(self, *args, **kwargs)
        logging.debug("%s.%s => %s", classname, func.__name__, res)
        return res

    return inner



import os
import os.path
import shlex
import sys
from argparse import ArgumentParser, ArgumentTypeError, Namespace
from functools import wraps
from os import makedirs
from pathlib import Path
from typing import Any, Callable, Union


def get_args(argparser: ArgumentParser) -> Namespace:
    """get commandline arguments from std input instead"""

    from pprint import pprint

    from .stdio import confirm

    if len(sys.argv) > 1:
        return argparser.parse_args()

    print("stdin")
    args = []

    for action in argparser._actions:
        pprint(action)

    for k, v in argparser._option_string_actions.items():
        nargs = v.nargs if v.nargs is not None else 1

        if nargs == 0:
            if confirm(str(k)):
                args.append(k)
        else:
            instr = input(f"{k} ({nargs}): ")  # separate multiple value by whitespace, quoting supported.
            args.append(k)
            args.append(instr)

    args = shlex.split(" ".join(args))
    pprint(args)
    return argparser.parse_args(args)


def arg_to_path(func: Callable[[Path], Path]) -> Callable:
    @wraps(func)
    def inner(path):
        return func(Path(path))

    return inner


def multiple_of(divisor: int) -> Callable[[str], int]:
    from builtins import int as builtin_int

    """ This function is called 'int' so that argparse can show a nicer error message
        in case input cannot be cast to int:
        error: argument --multiple: invalid int value: 'a'
    """

    def int(s: str) -> builtin_int:
        number = builtin_int(s)

        if number % divisor != 0:
            msg = f"{s} is not clearly divisible by {divisor}"
            raise ArgumentTypeError(msg)

        return number

    return int


def in_range(start: int, stop: int, step: int = 1) -> Callable[[str], int]:
    from builtins import int as builtin_int

    def int(s: str) -> builtin_int:  # see: multiple_of()
        number = builtin_int(s)

        r = range(start, stop, step)
        if number not in r:
            msg = f"{s} is not in {r}"
            raise ArgumentTypeError(msg)

        return number

    return int


def between(start: float, stop: float) -> Callable[[str], float]:
    from builtins import float as builtin_float

    def float(s: str) -> builtin_float:
        number = builtin_float(s)

        if not (start <= number < stop):
            msg = f"{s} is not in between {start} and {stop}"
            raise ArgumentTypeError(msg)

        return number

    return float


def suffix(s: str) -> str:
    """Checks if `s` is a valid suffix with a leading dot."""

    if len(s) <= 1 or not s.startswith("."):
        msg = f"{s} is not a valid suffix. It must start with a dot."
        raise ArgumentTypeError(msg)

    return s


def suffix_raw(s: str) -> str:
    """Checks if `s` is a valid suffix which starts with a dot,
    but returns it without the dot.
    """

    if len(s) <= 1 or not s.startswith("."):
        msg = f"{s} is not a valid suffix. It must start with a dot."
        raise ArgumentTypeError(msg)

    return s[1:]


def lowercase(s: str) -> str:
    """Converts argument to lowercase."""

    return s.lower()


def suffix_lower(s: str) -> str:
    return lowercase(suffix(s))


def suffix_lower_raw(s: str) -> str:
    return lowercase(suffix_raw(s))


@arg_to_path
def existing_path(path: Path) -> Path:
    """Checks if a path exists."""

    if not path.exists():
        msg = f"{path} does not exist"
        raise ArgumentTypeError(msg)

    return path


@arg_to_path
def new_path(path: Path) -> Path:
    """Checks if a path exists."""

    if path.exists():
        msg = f"{path} already exists"
        raise ArgumentTypeError(msg)

    return path


@arg_to_path
def is_dir(path: Path) -> Path:
    """Checks if a path is an actual directory"""

    if not path.is_dir():
        msg = f"{path} is not a directory"
        raise ArgumentTypeError(msg)

    return path


@arg_to_path
def abs_path(path: Path) -> Path:
    """Checks if a path is an actual directory"""

    return path.resolve()


def is_rel_path(path: str) -> str:
    """Checks if a path is relative"""

    if os.path.isabs(path):
        msg = f"{path} is absolute"
        raise ArgumentTypeError(msg)

    return path


@arg_to_path
def is_file(path: Path) -> Path:
    """Checks if a path is an actual file"""

    if not path.is_file():
        msg = f"{path} is not a file"
        raise ArgumentTypeError(msg)

    return path


@arg_to_path
def future_file(path: Path) -> Path:
    """Tests if file can be created to catch errors early.
    Checks if directory is writeable and file does not exist yet.
    """

    if path.parent and not os.access(str(path.parent), os.W_OK):
        msg = f"cannot access directory {path.parent}"
        raise ArgumentTypeError(msg)
    if path.is_file():
        msg = f"file {path} already exists"
        raise ArgumentTypeError(msg)
    return path


@arg_to_path
def out_dir(path: Path) -> Path:
    """Tests if `path` is a directory. If not it tries to create one."""

    if not path.is_dir():
        try:
            makedirs(path)
        except OSError:
            msg = f"Error: '{path}' is not a valid directory."
            raise ArgumentTypeError(msg)

    return path


@arg_to_path
def empty_dir(dirname: Path) -> Path:
    """tests if directory is empty"""

    from .iter import is_empty

    with os.scandir(dirname) as it:
        if not is_empty(it):
            msg = f"directory {dirname} is not empty"
            raise ArgumentTypeError(msg)

    return dirname


@arg_to_path
def empty_dir_create(dirname: Path) -> Path:
    """tests if directory is empty"""

    from .iter import is_empty

    if dirname.exists():
        with os.scandir(dirname) as it:
            if not is_empty(it):
                msg = f"directory {dirname} is not empty"
                raise ArgumentTypeError(msg)
    else:
        dirname.mkdir(parents=True)

    return dirname


def json_file(path: Union[str, Path]) -> Any:
    from json import JSONDecodeError

    from .json import read_json

    try:
        return read_json(path)
    except JSONDecodeError as e:
        raise ArgumentTypeError(f"JSONDecodeError: {e}")


def base64(s: str) -> bytes:
    """Checks if `s` is a valid base64 and decodes it."""
    import binascii
    from base64 import b64decode

    try:
        return b64decode(s, validate=True)
    except binascii.Error:
        msg = f"{s} is not valid base64"
        raise ArgumentTypeError(msg)


def ascii(s: str) -> str:
    """Checks if `s` is a valid ascii."""
    try:
        s.encode("ascii")
    except UnicodeEncodeError:
        msg = f"{s} is not valid ascii"
        raise ArgumentTypeError(msg)

    return s


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--str")
    parser.add_argument("--required", action="store_true")
    parser.add_argument("--add", nargs="+")
    print(get_args(parser))

    """
    parser = ArgumentParser()
    parser.add_argument('--indir', type=is_dir)
    parser.add_argument('--infile', type=is_file)
    parser.add_argument('--outfile', type=future_file)
    parser.add_argument('--outdir', type=empty_dir)
    args = parser.parse_args()
    print(args)
    """



from datetime import datetime
from datetime import time as dtime
from datetime import timedelta, timezone
from typing import Optional, overload

utcmin = datetime.min.replace(tzinfo=timezone.utc)
utcmax = datetime.max.replace(tzinfo=timezone.utc)


def current_local_timezone() -> timezone:
    """Returns a local fixed offset timezone.
    Warning: This timezone is only valid at the time of creation.
    Ie. if called in the summer for a local timezone which follows DST,
    and applied on a winter datetime, the result will be wrong.
    For a correct implementation use something like `dateutil.tz.tzlocal()` or `tzlocal.get_localzone()`.
    """

    return datetime.now(timezone.utc).astimezone().tzinfo


def is_aware(dt: datetime) -> bool:
    """Test if a datetime object is aware or naive."""

    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None


def naive_to_aware(dt, tzinfo=timezone.utc) -> datetime:
    if is_aware(dt):
        return dt
    else:
        return dt.replace(tzinfo=tzinfo)


def now(aslocal: bool = False) -> datetime:
    """Returns the current datetime as timezone aware object in
    UTC timezone if `aslocal=False` (the default)
    or local timezone if `aslocal=True`.
    """

    dt = datetime.now(timezone.utc)
    if aslocal:
        dt = dt.astimezone(None)
    return dt


def datetime_from_utc_timestamp(epoch: float, aslocal: bool = False) -> datetime:
    """Converts a UNIX epoch time in seconds to a timezone aware datetime.
    Negative values are supported and return a datetime counted backwards
    from 1970-01-01 UTC.
    `aslocal=True` doesn't work for negative values on Windows.
    """

    # don't use `fromtimestamp`, to have better cross platform support
    dt = datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(seconds=epoch)

    if aslocal:
        dt = dt.astimezone(None)
    return dt


def datetime_from_utc_timestamp_ms(epoch: int, aslocal: bool = False) -> datetime:
    """Converts a UNIX epoch time in milliseconds to a timezone aware datetime.
    Negative values are supported and return a datetime counted backwards
    from 1970-01-01 UTC.
    `aslocal=True` doesn't work for negative values on Windows.
    """

    seconds, ms = divmod(epoch, 1000)

    dt = datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(seconds=seconds, milliseconds=ms)

    if aslocal:
        dt = dt.astimezone(None)

    return dt


def datetime_from_utc_timestamp_ns(epoch: int, aslocal: bool = False) -> datetime:
    """Converts a UNIX epoch time in nano seconds to a timezone aware datetime.
    Negative values are supported and return a datetime counted backwards
    from 1970-01-01 UTC.
    `aslocal=True` doesn't work for negative values on Windows.
    """

    seconds, ns = divmod(epoch, 1000000000)

    # don't use `fromtimestamp`, to have better cross platform support
    dt = datetime(1970, 1, 1, tzinfo=timezone.utc) + timedelta(seconds=seconds, microseconds=ns // 1000)

    if aslocal:
        dt = dt.astimezone(None)

    return dt


@overload
def between(dt: datetime, start: Optional[datetime], end: Optional[datetime]) -> bool:
    pass


@overload
def between(dt: dtime, start: Optional[dtime], end: Optional[dtime]) -> bool:
    pass


def between(dt, start=None, end=None):
    """Tests if `dt` is in-between `start` and `end` (inclusive and optionally open ended).

    If the parameters are datetimes, they all most be either offset-aware or native.
    If `start` and `end` are `time` objects (without dates),
    then `end` can come before `start` to specify ranges which overlap two days.

    If `start` equals `end`, return `True`.
    """

    if start and end:
        if start < end:
            return start <= dt and dt <= end
        else:
            return start <= dt or dt <= end

    if start and dt < start:
        return False
    if end and dt > end:
        return False
    return True



import importlib
import logging
import pickle  # nosec
from typing import Any, Iterable, Iterator, Optional, Tuple

from ._files import PathType
from .atomic import sopen
from .file import copen

logger = logging.getLogger(__name__)


def read_pickle(path: PathType) -> Any:
    """Read pickle file from `path`.
    Warning: All usual security consideration regarding the pickle module still apply.
    """

    with copen(path, "rb") as fr:
        return pickle.load(fr)  # nosec


def write_pickle(result: Any, path: PathType, protocol: Optional[int] = None, safe: bool = False) -> None:
    """Write `result` to `path` using pickle serialization.

    `protocol': pickle protocol version
    `safe`: if True, don't overwrite original file in case any error occurs
    """

    with sopen(path, "wb", safe=safe) as fw:
        pickle.dump(result, fw, protocol=protocol)


def read_iter(path: PathType) -> Iterator[Any]:
    """Read pickled iterable from `path`.
    Warning: All usual security consideration regarding the pickle module still apply.
    """

    with copen(path, "rb") as fr:
        unpickler = pickle.Unpickler(fr)  # nosec
        while fr.peek(1):
            yield unpickler.load()


def write_iter(it: Iterable[Any], path: PathType, protocol: Optional[int] = None, safe: bool = False) -> Iterator[Any]:
    """Write iterable `it` to `path` using pickle serialization. This uses much less memory than
            writing a full list at once.
    Read back using `read_iter()`. If `safe` is True, the original file is not overwritten
            if any error occurs.
    This is a generator which yields the values read from `it`. So it must be consumed
            to actually write anything to disk.
    """

    with sopen(path, "wb", safe=safe) as fw:
        pickler = pickle.Pickler(fw, protocol=protocol)
        for result in it:
            pickler.dump(result)
            yield result


def key_to_hash(key: Any, protocol: Optional[int] = None) -> str:
    from hashlib import md5

    binary = pickle.dumps(key, protocol=protocol)
    return md5(binary).hexdigest()  # nosec


def unpickle(path: PathType, requirements: Iterable[Tuple[str, Optional[str]]] = ()) -> Any:
    """Can be used to unpickle objects when normal unpickling fails to import some dependencies correctly.
    path: Path to the pickled file.
    requirements: Iterable of (module, package) tuples to be imported.

    Consider this scenario:
    ```python
    import pickle
    class asd:
            def qwe():
                    pass
    a = asd()
    with open("asd.p", "wb") as fw:
            pickle.dump(a, fw)

    with open("asd.p", "rb") as fr:
            a = pickle.load(fr)  # everything ok here
    ```

    Now the class definition is moved out of this file into another file `qwe.py`. Then
    ```python
    import pickle
    with open("asd.p", "rb") as fr:
            a = pickle.load(fr)  # AttributeError: Can't get attribute 'asd' on <module '__main__' (built-in)>
    ```

    Now unpickle can be used to fix this problem:
    ```python
    from genutility.pickle import unpickle
    a = unpickle("asd.p", [("qwe", "asd")])
    ```

    Warning: Only use on safe inputs.
    """

    import __main__

    for module, package in requirements:
        mod = importlib.import_module(module, package)
        if package:
            name = package
            type = getattr(mod, package)
        else:
            name = module
            type = mod

        setattr(__main__, name, type)

    with copen(path, "rb") as fr:
        return pickle.load(fr)  # nosec



import logging
import signal
from types import FrameType
from typing import Any, Callable, Iterable, Optional, TypeVar

T = TypeVar("T")

logger = logging.getLogger(__name__)


class HandleKeyboardInterrupt:
    """Context manager to handle ctrl-c (keyboard interrupts)

    with HandleKeyboardInterrupt():
            do_a()
            do_b()

    Any KeyboardInterrupt received within the context manager will be caught and possibly
    re-raised at the end of the block. It can be used to make sure that sensitive code is not
    carelessly interrupted in the middle of execution by the user. The process could still be
    kill though, so this method should only be used as a convenience function, not to implement
    atomic transactions.

    https://stackoverflow.com/questions/842557/how-to-prevent-a-block-of-code-from-being-interrupted-by-keyboardinterrupt-in-py#comment81343976_21919644
    "this code may call third-party exception handlers in threads other than the main thread, which CPython never does"
    so be careful with threads
    """

    def __init__(self, raise_after: bool = True, delay: int = -1) -> None:
        """If `raise_after` is True, the KeyboardInterrupt will be re-raised when leaving the context.
        `delay` specifies the maximum number of times the signal is caught.
        """

        self.init_raise_after = raise_after
        self.init_delay = delay

    def __enter__(self) -> None:
        self.raise_after = self.init_raise_after  # pylint: disable=attribute-defined-outside-init
        self.delay = self.init_delay  # pylint: disable=attribute-defined-outside-init
        self.signal_received: Optional[tuple] = None
        self.old_handler = signal.signal(signal.SIGINT, self.handler)

    def handler(self, sig: signal.Signals, frame: FrameType) -> None:
        if self.delay == 0:
            self.raise_after = False
            self.old_handler(sig, frame)  # type: ignore # see cast above
        else:
            self.delay -= 1
            self.signal_received = (sig, frame)
            logger.debug("SIGINT received. Delaying KeyboardInterrupt.")

    def __exit__(self, exc_type, exc_value, traceback):
        signal.signal(signal.SIGINT, self.old_handler)
        if self.signal_received and self.raise_after:
            self.old_handler(*self.signal_received)


def safe_for_loop(
    it: Iterable[T], body_func: Callable[[T], bool], else_func: Optional[Callable[[], Any]] = None
) -> None:
    """This function makes sure that `body_func(element)` is called each time an element is retrieved from the
    iterable `it`, even if a `KeyboardInterrupt` is signaled in-between.
    After `body_func` is called, the KeyboardInterrupt will be re-raised and can be caught in
    calling function. `body_func` can optionally return `True` to exit the loop early.
    If the loop finishes, `else_func` will be called if provided.
    """

    it = iter(it)
    finished = False
    Uninterrupted = HandleKeyboardInterrupt(True)
    while True:
        with Uninterrupted:
            try:
                chunk = next(it)
            except StopIteration:
                finished = True
                break
            if body_func(chunk):
                break
    if finished and else_func:  # no break
        with Uninterrupted:
            else_func()



from typing import Iterable, Iterator

import numpy as np

from .image import image_block_histogram, image_histogram
from .iter import pairwise
from .numpy import histogram_correlation


def scene_change_detection_histogram_correlation(images: Iterable[np.ndarray]) -> Iterator[np.ndarray]:
    """Based on: 'Histogram Correlation for Video Scene Change Detection'. Algorithm described
    in paper is incomplete. It only mentions 'almost constant', 'not constant' and 'change is sharp'
    behaviour of the correlation. A real implementation would need to define a sliding window
    based derivative and threshold values for said derivatives.
    """

    it = iter(images)
    hist_rf = image_histogram(np.array(next(it)))

    for image in it:
        hist_i = image_histogram(np.array(image))
        yield histogram_correlation(hist_rf, hist_i)


def scene_change_detection_block_histogram(images: Iterator[np.ndarray]) -> Iterator[np.ndarray]:
    it = iter(images)

    def get_means_of_block_histograms(arr: np.ndarray) -> np.ndarray:
        hist = image_block_histogram(arr, 16, 16)
        # print(hist.shape)
        # print(hist.tolist())
        means = np.mean(hist, axis=-1)
        print(means)
        return means

    for means1, means2 in pairwise(map(get_means_of_block_histograms, it)):
        # difference of means
        diffs = np.abs(means1 - means2)
        yield diffs


"""
def proc(hist: np.ndarray, lambda_: Union[float, int] = 200):
    delta = 0.5
    DS = np.sqrt(2.0) / 2.0 * np.sum(np.abs(hist - hist.T), axis=(-2, -1))
    B = DS > lambda_
    BC = np.mean(unpack(B, 2, 2), axis=-1) > delta  # noqa

    raise RuntimeError("Unfinished")
"""


def test_scene_change_detection_histogram_correlation():
    from .cv import iter_video

    images = iter_video()
    for score in scene_change_detection_histogram_correlation(images):
        print(score)


def test_scene_change_detection_block_histogram():
    from .cv import iter_video

    images = iter_video()
    for diffs in scene_change_detection_block_histogram(images):
        print(diffs.tolist())


if __name__ == "__main__":
    # test_image_histogram_gray()
    test_scene_change_detection_histogram_correlation()
    test_scene_change_detection_block_histogram()



from contextlib import ExitStack
from os import PathLike
from pathlib import Path
from typing import Iterator, Optional, Union

from PyPDF2 import PdfReader, PdfWriter

PathType = Union[PathLike, str]


def join_pdfs_in_folder(
    path_in: Path,
    file_out: PathType,
    overwrite: bool = False,
    rotate: Optional[int] = None,
    password: Optional[Union[str, bytes]] = None,
) -> None:
    """Join all PDFs in `path_in` and write to `file_out`.
    `overwrite`: Overwrite existing output files
    `rotate`: Optionally rotate pages in 90 degree increments
    `password`: Optionally specify input file password
    """

    if not path_in.exists():
        raise FileNotFoundError(path_in)

    with ExitStack() as stack:
        paths = sorted(path_in.glob("*.pdf"))
        writer = PdfWriter()

        for path in paths:
            if path.is_file():
                fr = stack.enter_context(path.open("rb"))
                reader = PdfReader(fr, strict=True, password=password)
                for page in reader.pages:
                    if rotate is not None:
                        page = page.rotate(rotate)
                    writer.add_page(page)

        if overwrite:
            mode = "w"
        else:
            mode = "x"

        with open(file_out, mode + "b") as fw:
            writer.write_stream(fw)


def iter_pdf_text(path: str) -> Iterator[str]:
    with open(path, "rb") as fr:
        pdf = PdfReader(fr)
        for page in pdf.pages:
            yield page.extract_text()


def _read_pdf_pdfminer(path: str) -> str:
    from pdfminer.high_level import extract_text

    return extract_text(path)


def _read_pdf_tika(path: str) -> str:
    from tika import parser

    raw = parser.from_file(path)

    # return raw["metadata"]
    return raw["content"]


def read_pdf(path: str, engine: str = "pdfminer") -> str:
    try:
        func = {
            "pdfminer": _read_pdf_pdfminer,
            "tika": _read_pdf_tika,
        }[engine]
    except KeyError:
        raise ValueError(f"Engine {engine} doesn't exist")

    return func(path)


if __name__ == "__main__":
    from argparse import ArgumentParser

    from genutility.args import is_dir

    parser = ArgumentParser(description="Merge pdf files in directory into one file.")
    parser.add_argument("dir", type=is_dir, help="input directory")
    parser.add_argument("out", type=Path, help="output file path")
    parser.add_argument("--rotate", type=int, help="Rotate in 90 degree increments")
    parser.add_argument("--overwrite", action="store_true", help="Overwrite existing output files")
    parser.add_argument("--password", type=str, help="Input file password")
    args = parser.parse_args()

    join_pdfs_in_folder(args.dir, args.out, args.overwrite, args.rotate, args.password)



from typing import TYPE_CHECKING, Iterator, Union

import cv2
import numpy as np

from videofile import CvVideo

if TYPE_CHECKING:
    import wx  # noqa: F401


def wx_to_cv_image(wximage: "wx.Image") -> np.ndarray:
    buf = wximage.GetDataBuffer()
    arr = np.asarray(buf, dtype=np.uint8)

    return np.reshape(arr, (wximage.Height, wximage.Width, 3))[:, :, ::-1]  # RGB => BGR


def grayscale(cvimg: np.ndarray) -> np.ndarray:
    return cv2.cvtColor(cvimg, cv2.COLOR_RGB2GRAY)


def iter_video(input: Union[str, int] = 0, show: bool = False) -> Iterator[np.ndarray]:
    with CvVideo(input) as video:
        if show:
            for _time, image in video.show():
                yield image
        else:
            for _time, image in video.iterall(native=True):
                yield image



from operator import itemgetter
from typing import Callable, Iterator, Optional, Set, Tuple, TypeVar, Union

T = TypeVar("T")


class RoutingObject:
    def __init__(self, value: T) -> None:
        self.value = value

        self.subtree: Optional["NodeType"] = None
        self.distance_to_parent: Optional[float] = None
        self.covering_radius: Optional[float] = None

    def __repr__(self) -> str:
        return f"RoutingObject({self.value}, radius={self.covering_radius}, tree={self.subtree})"

    def set_subtree(self, node) -> None:
        self.subtree = node
        node.parent_object = self


class LeafObject:
    def __init__(self, value: T) -> None:
        self.value = value

        self.distance_to_parent: Optional[float] = None

    def __repr__(self):
        return f"LeafObject({self.value})"

    def to_routing_object(self):
        return RoutingObject(self.value)


def isid(obj):
    if obj is None:
        return None
    else:
        return id(obj)


class MNode:
    maxsize = 4

    def __init__(
        self, distance_func: Callable, parent_node: Optional["NodeType"], parent_object: Optional["ObjectType"]
    ) -> None:
        self.distance_func = distance_func
        self.parent_node = parent_node
        self.parent_object = parent_object

    @property
    def is_root(self) -> bool:
        return self.parent_node is None

    @property
    def is_full(self) -> bool:
        return len(self.objects) >= self.maxsize

    def add(self, obj: "ObjectType") -> None:
        self.objects.add(obj)
        if isinstance(obj, RoutingObject):
            obj.subtree.parent_node = self

        if not self.is_root:
            obj.distance_to_parent = self.distance_func(obj.value, self.parent_object.value)

    def replace(self, o1: "ObjectType", o2: "ObjectType") -> None:
        self.objects.remove(o1)
        self.objects.add(o2)

    def __repr__(self):
        return repr(
            {
                "parent_node": isid(self.parent_node),
                "parent_object": isid(self.parent_object),
                "objects": self.objects,
            }
        )


class InternalNode(MNode):
    def __init__(
        self, distance_func: Callable, parent_node: Optional["NodeType"], parent_object: Optional["ObjectType"]
    ) -> None:
        MNode.__init__(self, distance_func, parent_node, parent_object)
        self.objects: Set[RoutingObject] = set()

    def set_objects(self, objects: Set[RoutingObject]) -> None:
        self.objects = objects
        op = self.parent_object

        for o in objects:
            assert o.subtree.parent_object is o
            o.subtree.parent_node = self

        if not self.is_root:
            for o in objects:
                o.distance_to_parent = self.distance_func(op.value, o.value)

            if objects:
                op.covering_radius = max(o.distance_to_parent + o.covering_radius for o in objects)
            else:
                op.covering_radius = 0

    def update_objects(self):
        for o in self.objects:
            assert o.subtree.parent_object is o
            assert o.subtree.parent_node is self

        if not self.is_root:
            op = self.parent_object

            for o in self.objects:
                o.distance_to_parent = self.distance_func(op.value, o.value)

            op.covering_radius = max(o.distance_to_parent + o.covering_radius for o in self.objects)


class LeafNode(MNode):
    def __init__(
        self, distance_func: Callable, parent_node: Optional["NodeType"], parent_object: Optional["ObjectType"]
    ) -> None:
        MNode.__init__(self, distance_func, parent_node, parent_object)
        self.objects: Set[LeafObject] = set()

    def set_objects(self, objects: Set[LeafObject]) -> None:
        self.objects = objects
        op = self.parent_object

        if not self.is_root:
            for o in objects:
                o.distance_to_parent = self.distance_func(op.value, o.value)

            if objects:
                op.covering_radius = max(o.distance_to_parent for o in objects)
            else:
                op.covering_radius = 0

    def update_objects(self):
        if not self.is_root:
            op = self.parent_object

            for o in self.objects:
                o.distance_to_parent = self.distance_func(op.value, o.value)

            op.covering_radius = max(o.distance_to_parent for o in self.objects)


ObjectType = Union[RoutingObject, LeafObject]
NodeType = Union[InternalNode, LeafNode]

from random import sample


def _promote_random(distance_func: Callable, objects: Set[ObjectType]) -> Tuple[ObjectType, ObjectType]:
    a, b = sample(objects, 2)
    return a, b


def _partition_generalized_hyperplane(
    distance_func: Callable, objects: Set[ObjectType], o1: ObjectType, o2: ObjectType
) -> Tuple[Set[ObjectType], Set[ObjectType]]:
    a: Set[ObjectType] = set()
    b: Set[ObjectType] = set()

    for o in objects:
        d1 = distance_func(o.value, o1.value)
        d2 = distance_func(o.value, o2.value)

        if d1 <= d2:
            a.add(o)
        else:
            b.add(o)

    return a, b


class MTree:
    """See: M-tree: An Efficient Access Method for Similarity Search in Metric Spaces (1997)"""

    def __init__(
        self,
        distance_func: Callable[[ObjectType, ObjectType], float],
        promote: Optional[str] = None,
        partition: Optional[str] = None,
        does_not_work_yet: Optional[str] = None,
    ) -> None:
        if does_not_work_yet != "OK":
            raise RuntimeError("MTree is work in progress")

        self.distance_func = distance_func

        self.root = LeafNode(self.distance_func, None, None)

        self._promote = {
            "random": _promote_random,
        }[promote or "random"]
        self._partition = {
            "generalized-hyperplane": _partition_generalized_hyperplane,
        }[partition or "generalized-hyperplane"]

    def __repr__(self):
        return f"<MTree: {self.root!r}>"

    def _keys(self, node):
        if isinstance(node, InternalNode):
            for ro in node.objects:
                yield from self._keys(ro.subtree)
        else:
            for lo in node.objects:
                yield lo.value

    def _split(self, node: NodeType, obj: ObjectType) -> None:
        # if not node.is_root:
        op = node.parent_object
        np = node.parent_node

        n_new = type(node)(self.distance_func, np, op)  # same type like existing
        all_objects = node.objects | {obj}
        o1, o2 = self._promote(self.distance_func, all_objects)
        os1, os2 = self._partition(self.distance_func, all_objects, o1, o2)

        if isinstance(node, LeafNode):
            o1 = o1.to_routing_object()
            o2 = o2.to_routing_object()

        o1.set_subtree(node)
        o2.set_subtree(n_new)

        if node.is_root:
            # print("is_root")
            self.root = InternalNode(self.distance_func, None, None)
            self.root.add(o1)
            self.root.add(o2)

            node.set_objects(os1)  # recalculates distances
            n_new.set_objects(os2)

        else:
            np.replace(op, o1)
            if np.is_full:
                # print("is_full")
                self._split(np, o2)
            else:
                # print("is not full")
                np.add(o2)
                np.update_objects()

    def _add(self, node: NodeType, obj: LeafObject) -> None:
        if not isinstance(node, LeafNode):
            # print("not leaf", node)
            distances = [self.distance_func(ro.value, obj.value) for ro in node.objects]

            n_in = [(ro, d) for ro, d in zip(node.objects, distances) if d <= ro.covering_radius]
            if n_in:
                found = min(n_in, key=itemgetter(1))[0]
            else:
                it = ((ro, d - ro.covering_radius, d) for ro, d in zip(node.objects, distances))
                min_entry = min(it, key=itemgetter(1))
                found = min_entry[0]
                found.covering_radius = min_entry[2]

            self._add(found.subtree, obj)
        else:
            if not node.is_full:
                # print("not full", "LeafNode", node)
                node.add(obj)
            else:
                # print("is full", "LeafNode", node)
                self._split(node, obj)

    def _find(self, node, value, radius):
        assert not node.is_root

        distance_parent_to_value = self.distance_func(node.parent_object.value, value)

        if not isinstance(node, LeafNode):
            for ro in node.objects:
                if abs(distance_parent_to_value - ro.distance_to_parent) <= radius + ro.covering_radius:
                    if self.distance_func(ro.value, value) <= radius + ro.covering_radius:
                        yield from self._find(ro.subtree, value, radius)
        else:
            for lo in node.objects:
                if abs(distance_parent_to_value - lo.distance_to_parent) <= radius:
                    if self.distance_func(lo.value, value) <= radius:
                        yield lo.value

    def keys(self):
        yield from self._keys(self.root)

    def add(self, value: T) -> None:
        self._add(self.root, LeafObject(value))

    def find(self, value: T, radius: float) -> Iterator[T]:
        node = self.root

        if not isinstance(node, LeafNode):
            for ro in node.objects:
                if self.distance_func(ro.value, value) <= radius + ro.covering_radius:
                    yield from self._find(ro.subtree, value, radius)
        else:
            for lo in node.objects:
                if self.distance_func(lo.value, value) <= radius:
                    yield lo.value


def len_dist(s1, s2):
    return abs(len(s1) - len(s2))


def vals(start: int, end: int) -> Set[str]:
    return {str(i) * i for i in range(start, end + 1)}


from unittest import TestCase


class MTreeTests(TestCase):
    def test_mtree_one_layer(self):
        values = vals(1, 4)

        mt = MTree(len_dist, does_not_work_yet="OK")
        for i in values:
            mt.add(i)

        assert set(mt.find("333", 1)) == vals(2, 4)
        assert set(mt.keys()) == values

    def test_mtree_two_layer(self):
        values = vals(1, 5)

        mt = MTree(len_dist, does_not_work_yet="OK")
        for i in values:
            mt.add(i)

        assert set(mt.keys()) == values

    def test_mtree_three_layer(self):
        values = vals(1, 9)

        mt = MTree(len_dist, does_not_work_yet="OK")
        for i in values:
            mt.add(i)

        assert set(mt.find("333", 1)) == vals(2, 4)
        assert set(mt.keys()) == values


if __name__ == "__main__":
    import unittest

    unittest.main()



from typing import Any

import bs4


def find_one(node: bs4.element.Tag, *args: Any, **kwargs: Any) -> bs4.element.Tag:
    nodes = node.find_all(*args, **kwargs)

    if len(nodes) != 1:
        raise ValueError(f"Found {len(nodes)} nodes for: {args} {kwargs}")

    return nodes[0]


def find_zero_or_one(node: bs4.element.Tag, *args: Any, **kwargs: Any) -> bs4.element.Tag:
    nodes = node.find_all(*args, **kwargs)

    if len(nodes) == 0:
        return None
    elif len(nodes) == 1:
        return nodes[0]
    else:
        raise ValueError(f"Found {len(nodes)} nodes for: {args} {kwargs}")



import logging

import numpy as np
from PIL import Image, ImageFilter

# from .numba import opjit
from .numpy import rgb_to_hsi, rgb_to_ycbcr, unblock

# fingerprinting aka perceptual hashing


def phash_antijpeg(image: Image.Image) -> np.ndarray:
    """Source: An Anti-JPEG Compression Image Perceptual Hashing Algorithm
    `image` is a RGB pillow image.
    """

    raise NotImplementedError


def hu_moments(channels: np.ndarray) -> np.ndarray:
    """Calculates all Hu invariant image moments for all channels separately.
    Input array must be of shape [width, height, channels]
    Returns shape [moments, channels]
    """

    # pre-calculate matrices
    n, m, _ = channels.shape
    coords_x, coords_y = np.meshgrid(np.arange(m), np.arange(n))
    coords_x = np.expand_dims(coords_x, axis=-1)  # for batch input, some change is needed here
    coords_y = np.expand_dims(coords_y, axis=-1)  # for batch input, some change is needed here

    def M(p, q):
        return np.sum(coords_x**p * coords_y**q * channels, axis=(-2, -3))

    def mu(p, q, xb, yb):
        return np.sum((coords_x - xb) ** p * (coords_y - yb) ** q * channels, axis=(-2, -3))

    def eta(p, q, xb, yb, mu00):
        gamma = (p + q) / 2 + 1
        return mu(p, q, xb, yb) / mu00**gamma

    def loop():
        M00 = M(0, 0)
        if not np.all(M00 > 0.0):
            logging.error("M00: %s", M00)
            raise ValueError("Failed to calculate moments. Single color pictures are not supported yet.")

        M10 = M(1, 0)
        M01 = M(0, 1)
        xb = M10 / M00
        yb = M01 / M00

        mu00 = mu(0, 0, xb, yb)

        eta20 = eta(2, 0, xb, yb, mu00)
        eta02 = eta(0, 2, xb, yb, mu00)
        eta11 = eta(1, 1, xb, yb, mu00)
        eta30 = eta(3, 0, xb, yb, mu00)
        eta12 = eta(1, 2, xb, yb, mu00)
        eta21 = eta(2, 1, xb, yb, mu00)
        eta03 = eta(0, 3, xb, yb, mu00)

        phi1 = eta20 + eta02
        phi2 = (eta20 - eta02) ** 2 + 4 * eta11**2
        phi3 = (eta30 - 3 * eta12) ** 2 + (3 * eta21 - eta03) ** 2
        phi4 = (eta30 + eta12) ** 2 + (eta21 + eta03) ** 2
        phi5 = (eta30 - 3 * eta12) * (eta30 + eta12) * ((eta30 + eta12) ** 2 - 3 * (eta21 + eta03) ** 2) + (
            3 * eta21 - eta03
        ) * (eta21 + eta03) * (3 * (eta30 + eta12) ** 2 - (eta21 + eta03) ** 2)
        phi6 = (eta20 - eta02) * ((eta30 + eta12) ** 2 - (eta21 + eta03) ** 2) + 4 * eta11 * (eta30 + eta12) * (
            eta21 + eta03
        )
        phi7 = (3 * eta21 - eta03) * (eta30 + eta12) * ((eta30 + eta12) ** 2 - 3 * (eta21 + eta03) ** 2) - (
            eta30 - 3 * eta12
        ) * (eta21 + eta03) * (3 * (eta30 + eta12) ** 2 - (eta21 + eta03) ** 2)

        return np.array([phi1, phi2, phi3, phi4, phi5, phi6, phi7])

    return loop()


# @opjit() rgb_to_hsi and rgb_to_ycbcr not supported by numba
def phash_moments_array(arr: np.ndarray) -> np.ndarray:
    arr = arr / 255.0

    # convert colorspaces
    hsi = rgb_to_hsi(arr)
    ycbcr = rgb_to_ycbcr(arr)  # .astype(np.uint8)
    channels = np.concatenate([hsi, ycbcr], axis=-1)
    return np.concatenate(hu_moments(channels).T)


def phash_moments(image: Image.Image) -> np.ndarray:
    """Source: Perceptual Hashing for Color Images Using Invariant Moments
    `image` is a RGB pillow image. Results should be compared with L^2-Norm of difference vector.
    """

    if image.mode != "RGB":
        raise ValueError("Only RGB images are supported")

    # preprocessing
    image = image.resize((512, 512), Image.BICUBIC)
    image = image.filter(ImageFilter.GaussianBlur(3))
    image = np.array(image)
    return phash_moments_array(image)


def phash_blockmean_array(arr: np.ndarray, bits: int = 256) -> np.ndarray:
    """If bits is not a multiple of 8,
    the result will be zero padded from the right.
    """

    if len(arr.shape) != 2:
        raise ValueError("arr must be 2-dimensional")

    n = int(np.sqrt(bits))
    if n**2 != bits:
        raise ValueError("bits must be a square number")

    blocks = unblock(arr, n, n)
    means = np.mean(blocks, axis=-1)
    median = np.median(means)
    bools = means >= median
    return np.packbits(bools)


def phash_blockmean(image: Image.Image, bits: int = 256, x: int = 256) -> bytes:
    """Source: Block Mean Value Based Image Perceptual Hashing
    Method: 1
    Metric: 'Bit error rate' (normalized hamming distance)
    """

    image = image.convert("L").resize((x, x))
    image = np.array(image)
    return phash_blockmean_array(image, bits).tobytes()



import errno
import logging
import os
import os.path
import platform
import re
import shutil
import stat
from datetime import datetime
from fnmatch import fnmatch
from functools import partial
from itertools import chain, zip_longest
from operator import attrgetter
from pathlib import Path, PurePath
from typing import Any, Callable, Iterable, Iterator, List, Optional, Set, Tuple, Union

from typing_extensions import Self

from ._files import BaseDirEntry, MyDirEntryT, PathType, entrysuffix, to_dos_device_path
from .datetime import datetime_from_utc_timestamp
from .file import FILE_IO_BUFFER_SIZE, equal_files, iterfilelike
from .iter import is_empty
from .ops import logical_implication
from .os import _not_available, islink

EntryType = Union[Path, os.DirEntry]

logger = logging.getLogger(__name__)

ascii_control = {chr(i) for i in range(1, 32)}
ntfs_illegal_chars = {"\0", "/"}
fat_illegal_chars = {"\0", "/", "?", "<", ">", "\\", ":", "*", "|", '"', "^"}  # ^ really?

windows_ntfs_illegal_chars = ntfs_illegal_chars | {"\\", ":", "*", "?", '"', "<", ">", "|"} | ascii_control
windows_fat_illegal_chars = fat_illegal_chars

windows_illegal_chars = windows_ntfs_illegal_chars | windows_fat_illegal_chars
unix_illegal_chars = {"\0", "/"}
mac_illegal_chars = {"\0", "/", ":"}

windows_sep = r"\/"
linux_sep = r"/"
mac_sep = r"/:"

windows_reserved_names = {
    "CON",
    "PRN",
    "AUX",
    "NUL",
    "COM1",
    "COM2",
    "COM3",
    "COM4",
    "COM5",
    "COM6",
    "COM7",
    "COM8",
    "COM9",
    "LPT1",
    "LPT2",
    "LPT3",
    "LPT4",
    "LPT5",
    "LPT6",
    "LPT7",
    "LPT8",
    "LPT9",
}


class fileextensions:
    audio = ("wav", "mp3", "aac", "flac", "m4a", "m4b", "aiff", "ogg", "wma", "mka", "ac3", "dts")
    video = (
        "avi",
        "mp4",
        "m4v",
        "mkv",
        "mpg",
        "mpeg",
        "wmv",
        "mov",
        "flv",
        "f4v",
        "webm",
        "vob",
        "ogv",
        "m2ts",
        "rm",
        "rmvb",
        "asf",
        "3gp",
        "nsv",
        "divx",
        "ogm",
    )
    images = ("bmp", "jpg", "jpeg", "png", "gif", "tif", "tiff", "tga", "heif", "heic", "webp", "avif")

    archives = ("zip", "rar", "ace", "tar", "tgz", "tbz", "7z", "cab", "dmg", "wim")
    bytecode = ("pyc", "class")
    configuration = ("ini", "toml", "conf", "reg")
    database = ("cat", "sqlite", "db")
    documents = ("txt", "pdf", "doc", "docx", "ps", "rtf", "pgs")
    executables = ("exe", "dll", "so", "scr", "msi", "pyd", "sys", "cpl")  # native code
    image_archives = ("cbz", "cbr", "cb7", "cbt", "cba")
    scripts = ("bat", "ps1", "sh", "c", "cpp", "cs", "py", "php", "js", "lua", "pl")  # and source code files

    disc_images_binary = ("iso", "bin", "mdf", "img")
    disc_images_sidecar = ("cue", "mds", "ccd")
    disc_images = disc_images_binary + disc_images_sidecar

    game_images = ("gcn", "nds", "3ds", "wad", "xbx", "gba", "gb", "nes", "sfc", "n64", "z64")
    compressed = ("gz", "bz2", "lzma", "xz", "z")
    subtitles = ("srt", "sub", "idx", "sup", "ssa", "ass")
    scene = ("nfo", "sfv")

    windows = ("library-ms", "desklink", "lnk", "dmp")
    macos = ("crash",)
    os = windows + macos


class IllegalFilename(ValueError):
    pass


class WindowsIllegalFilename(IllegalFilename):
    pass


class FileProperties:
    __slots__ = ("relpath", "size", "isdir", "abspath", "id", "modtime", "hash")

    def __init__(
        self,
        relpath: Optional[str] = None,
        size: Optional[int] = None,
        isdir: Optional[bool] = None,
        abspath: Optional[str] = None,
        id: Optional[Any] = None,
        modtime: Optional[datetime] = None,
        hash: Optional[str] = None,
    ) -> None:
        if relpath is None and abspath is None:
            raise ValueError("Either `relpath` or `abspath` (or both) must be given")

        self.relpath = relpath
        self.size = size
        self.isdir = isdir
        self.abspath = abspath
        self.id = id
        self.modtime = modtime
        self.hash = hash

    @classmethod
    def keys(cls) -> Tuple[str, ...]:
        return cls.__slots__

    def values(self):
        return attrgetter(*self.__slots__)(self)

    def items(self):
        return zip(self.keys(), self.values())

    def __iter__(self) -> tuple:
        return iter(self.values())

    def __eq__(self, other: Any) -> bool:
        return self.values() == other.values()

    def __repr__(self) -> str:
        args = (f"{k}={v!r}" for k, v in self.items() if v is not None)
        return "FileProperties({})".format(", ".join(args))


class DirEntryStub:
    __slots__ = ("name", "path")

    def __init__(self, name: str, path: str) -> None:
        self.name = name
        self.path = path


class MyDirEntry(BaseDirEntry):
    __slots__ = ("_basepath", "_relpath", "follow")

    def __init__(self, entry: os.DirEntry) -> None:
        BaseDirEntry.__init__(self, entry)

        self._basepath: Optional[str] = None
        self._relpath: Optional[str] = None
        self.follow = True

    @property
    def basepath(self) -> str:
        if self._basepath is None:
            raise AttributeError("basepath not set")

        return self._basepath

    @basepath.setter
    def basepath(self, path: str) -> None:
        self._relpath = os.path.relpath(self.entry.path, path)
        self._basepath = path

    @property
    def relpath(self) -> str:
        if self._relpath is None:
            raise AttributeError("basepath not set")

        return self._relpath


if platform.system() == "Windows":

    def long_path_support(path: PathType) -> str:
        return to_dos_device_path(os.path.abspath(path))

else:

    def long_path_support(path: PathType) -> str:
        return os.fspath(path)


def mdatetime(path: PathType, aslocal: bool = False) -> datetime:
    """Returns the last modified date of `path`
    as a timezone aware datetime object.

    If `aslocal=True` it will be formatted as local time,
    and UTC otherwise (the default).
    """

    if isinstance(path, (Path, os.DirEntry)):
        mtime = path.stat().st_mtime
    else:
        mtime = os.stat(path).st_mtime

    return datetime_from_utc_timestamp(mtime, aslocal)


def rename(old: PathType, new: PathType) -> None:
    """Renames `old` to `new`. Fails if `new` already exists.
    This is the default behaviour of `os.rename` on Windows.
    This function should do the same cross-platform.
    It is however not race free.
    """

    # fixme: renaming a non-existing file in a non-existing folder yields a PermissionError not FileNotFoundError

    if os.path.exists(new):
        raise FileExistsError(new)

    os.renames(old, new)  # fixme: don't use rename*s*


def copy_file_generator(
    source: PathType, dest: PathType, buffer_size: int = FILE_IO_BUFFER_SIZE, overwrite_readonly: bool = False
) -> Iterator[None]:
    """Partial file is not deleted if exception gets raised anywhere."""

    try:
        os.makedirs(os.path.dirname(dest))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    if overwrite_readonly:
        try:
            make_writeable(dest)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    with open(source, "rb") as src, open(dest, "wb") as dst:
        for data in iterfilelike(src, chunk_size=buffer_size):
            dst.write(data)
            yield

    shutil.copystat(source, dest)


def st_mode_to_str(st_mode: int) -> str:
    if stat.S_ISDIR(st_mode):
        return "directory"
    elif stat.S_ISREG(st_mode):
        return "regular file"
    elif stat.S_ISCHR(st_mode):
        return "character special device file"
    elif stat.S_ISBLK(st_mode):
        return "block special device file"
    elif stat.S_ISFIFO(st_mode):
        return "named pipe"
    elif stat.S_ISLNK(st_mode):
        return "symbolic link"
    elif stat.S_ISSOCK(st_mode):
        return "socket"
    else:
        return "unknown"


def append_to_filename(path: PathType, s: str) -> str:
    """Add `s` to the filename given in `path`. It's added before the extension, not after."""

    root, ext = os.path.splitext(path)
    return root + s + ext


def scandir_error_log(entry: os.DirEntry, exception: Exception) -> None:
    if isinstance(exception, OSError) and exception.filename:
        logger.exception("Error processing path", exc_info=exception)
    else:
        logger.exception("Error in <%s>", entry.path, exc_info=exception)


def scandir_error_log_warning(entry: os.DirEntry, exception: Exception) -> None:
    if isinstance(exception, OSError) and exception.filename:
        logger.warning("Error processing path: %s", exception)
    else:
        logger.warning("Error in <%s>: %s", entry.path, exception)


def scandir_error_raise(entry: os.DirEntry, exception: Exception) -> None:
    raise exception


def scandir_error_ignore(entry: os.DirEntry, exception: Exception) -> None:
    pass


def _is_loop(rootentry: os.DirEntry, entry: os.DirEntry, is_link: bool, loops: Optional[Set[str]]) -> bool:
    _path: str

    if loops is not None:
        if is_link:
            _path = os.readlink(entry)

            if not os.path.isabs(_path):  # symlinks might be relative
                _path = os.path.join(rootentry.path, _path)
        else:
            _path = entry.path

        if _path in loops:
            return True
        else:
            loops.add(_path)

    return False


def _scandir_rec_skippable(
    rootentry: os.DirEntry,
    files: bool = True,
    others: bool = False,
    follow_symlinks: bool = True,
    prevent_loops_set: Optional[Set[str]] = None,
    errorfunc: Callable[[os.DirEntry, Exception], None] = scandir_error_raise,
) -> Iterator[MyDirEntry]:
    is_link = False

    try:
        with os.scandir(rootentry.path) as it:
            for entry in it:
                if files and entry.is_file(follow_symlinks=follow_symlinks):
                    yield MyDirEntry(entry)
                elif entry.is_dir(follow_symlinks=follow_symlinks):
                    if not follow_symlinks or prevent_loops_set is not None:
                        is_link = islink(entry)

                    se = MyDirEntry(entry)
                    yield se

                    if (
                        not follow_symlinks and is_link
                    ):  # must be a windows junction since `entry.is_symlink()` returns `False` for junctions
                        continue

                    if se.follow and not _is_loop(rootentry, entry, is_link, prevent_loops_set):
                        for e in _scandir_rec_skippable(
                            entry, files, others, follow_symlinks, prevent_loops_set, errorfunc
                        ):
                            yield e
                else:
                    if others:
                        yield MyDirEntry(entry)
    except OSError as e:
        errorfunc(rootentry, e)


def _scandir_rec(
    rootentry: os.DirEntry,
    files: bool = True,
    dirs: bool = False,
    others: bool = False,
    rec: bool = True,
    follow_symlinks: bool = True,
    prevent_loops_set: Optional[Set[str]] = None,
    errorfunc: Callable[[os.DirEntry, Exception], None] = scandir_error_raise,
) -> Iterator[os.DirEntry]:
    is_link = False

    try:
        with os.scandir(rootentry.path) as it:
            for entry in it:
                if files and entry.is_file(follow_symlinks=follow_symlinks):
                    yield entry
                elif entry.is_dir(follow_symlinks=follow_symlinks):
                    if not follow_symlinks or prevent_loops_set is not None:
                        is_link = islink(entry)

                    if dirs:
                        yield entry

                    if (
                        not follow_symlinks and is_link
                    ):  # must be a windows junction since `entry.is_symlink()` returns `False` for junctions
                        continue

                    if rec and not _is_loop(rootentry, entry, is_link, prevent_loops_set):
                        for e in _scandir_rec(
                            entry, files, dirs, others, rec, follow_symlinks, prevent_loops_set, errorfunc
                        ):
                            yield e
                else:
                    if others:
                        yield entry
    except OSError as e:
        errorfunc(rootentry, e)


def _get_entry_stub(path: str) -> os.DirEntry:
    return DirEntryStub(os.path.basename(path), long_path_support(path))


def scandir_rec(
    path: PathType,
    files: bool = True,
    dirs: bool = False,
    others: bool = False,
    rec: bool = True,
    follow_symlinks: bool = True,
    prevent_loops: bool = True,
    relative: bool = False,
    allow_skip: bool = False,
    errorfunc: Callable[[MyDirEntryT, Exception], None] = scandir_error_log,
) -> Iterator[MyDirEntryT]:
    r"""Variant of `os.scandir()` which recurses into subfolders.
    `path`: Relative or absolute filesystem path. On Windows this path is converted to a absolute device path <\\?\X:\path\file.exe> to support long paths.
    `files`: Yield files.
    `dirs`: Yield directories.
    `others`: Yield other entries which are neither files nor directories.
        If follow_symlinks is false, symlinks will be included with others,
        if follow_symlinks is true, they will be either in files or dirs.
        Junctions will always be part of dirs.
    `rec`: Recurse into subfolders.
    `follow_symlinks`: Follow symbolic links or directory junctions.
    `prevent_loops`: Directory links can lead to infinite recursion.
        If this is True, visited paths will be remembered and only followed once.
        Directories which are links will always be returned, but their contents only when they are encountered first.
        Symbolic links to files are always returned, which may lead to file duplicates, but not to loops.
        If `follow_symlinks` is False, this argument has no effect.
    `relative`: Yield `MyDirEntry` objects which expose a `relpath` attribute which contains the path as relativ to `path`
    `allow_skip`: Yield `MyDirEntry` objects which expose a `follow` attribute. If this is set to False for directories,
        they will not be followed.
    `errorfunc`: A callback function to handle errors. By default errors are logged as exceptions and ignored.


    CHANGE: previously junctions were not returned at all when follow_symlinks=False,
    """

    if not logical_implication(allow_skip, dirs and rec):
        raise ValueError("allow_skip implies dirs and rec")

    if isinstance(path, os.PathLike):
        path = os.fspath(path)

    if not follow_symlinks or not rec:
        prevent_loops = False

    entry = _get_entry_stub(path)
    prevent_loops_set: Optional[Set[str]] = set() if prevent_loops else None

    if not allow_skip:
        it = _scandir_rec(entry, files, dirs, others, rec, follow_symlinks, prevent_loops_set, errorfunc)
    else:
        it = _scandir_rec_skippable(entry, files, others, follow_symlinks, prevent_loops_set, errorfunc)

    if not relative:
        return it
    else:
        basepath = long_path_support(path)
        if not allow_skip:
            # entry is a `os.DirEntry`
            def modpathrelative(entry):
                entry = MyDirEntry(entry)
                entry.basepath = basepath
                return entry

        else:
            # entry is a `MyDirEntry`
            def modpathrelative(entry):
                entry.basepath = basepath
                return entry

        return map(modpathrelative, it)


def filter_recall(recall: bool = False) -> Callable[[MyDirEntryT], bool]:
    """Returns a functions to use with filter, which takes a direntry and returns False
    when `recall` is False and the file is only available online For example using OneDrive.
    """

    if os.name == "nt" and not recall:
        from .os_win import FileAttributes

        def inner(entry: MyDirEntryT) -> bool:
            if FileAttributes.RECALL_ON_DATA_ACCESS in FileAttributes(entry.stat().st_file_attributes):
                logger.info("Skipping online-only file `%s`", entry.path)
                return False
            return True

    else:

        def inner(entry: MyDirEntryT) -> bool:
            return True

    return inner


def scandir_rec_simple(
    path: str,
    files: bool = True,
    dirs: bool = False,
    others: bool = False,
    rec: bool = True,
    follow_symlinks: bool = True,
    prevent_loops: bool = True,
    errorfunc: Callable[[MyDirEntryT, Exception], None] = scandir_error_log_warning,
) -> Iterator[os.DirEntry]:
    entry = DirEntryStub(os.path.basename(path), path)
    prevent_loops_set: Optional[Set[str]] = set() if prevent_loops else None

    return _scandir_rec(entry, files, dirs, others, rec, follow_symlinks, prevent_loops_set, errorfunc)


def scandir_ext(
    path: PathType,
    include: Optional[Set[str]] = None,
    exclude: Optional[Set[str]] = None,
    rec: bool = True,
    follow_symlinks: bool = False,
    relative: bool = False,
    errorfunc: Callable = scandir_error_log,
) -> Iterator[MyDirEntryT]:
    if (include is not None) and (exclude is not None):
        raise ValueError("include and exclude cannot be used together")

    it = scandir_rec(
        path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks, relative=relative, errorfunc=errorfunc
    )
    if include is not None:
        for entry in it:
            if entrysuffix(entry).lower() in include:
                yield entry
    elif exclude is not None:
        for entry in it:
            if entrysuffix(entry).lower() not in exclude:
                yield entry
    else:
        yield from it


# fixme: benchmark and delete
def scandir_ext_2(path: str, extensions: Set[str], rec: bool = True) -> Iterator[Path]:
    if rec:
        paths = Path(long_path_support(path)).rglob("*")
    else:
        paths = Path(long_path_support(path)).glob("*")

    for p in paths:
        if p.suffix.lower() in extensions:
            yield p


def _scandir_depth(rootentry: os.DirEntry, depth: int, errorfunc: Callable):
    try:
        with os.scandir(rootentry.path) as it:
            for entry in it:
                if entry.is_dir():
                    yield depth, entry
                    yield from _scandir_depth(entry, depth + 1, errorfunc)

        with os.scandir(rootentry.path) as it:
            for entry in it:
                if entry.is_file():
                    yield depth, entry

    except OSError as e:
        errorfunc(rootentry, e)


def scandir_depth(
    path: str, depth: int = 0, onerror: Callable = scandir_error_log
) -> Iterator[Tuple[int, os.DirEntry]]:
    """Recursive version of `scandir` which yields the current path depth
    along with the os.DirEntry.
    Directories are returned first, then files.
    The order is arbitrary otherwise.
    """

    entry = DirEntryStub(
        os.path.basename(path), long_path_support(path)
    )  # for python 2 compat. and long filename support
    return _scandir_depth(entry, depth, onerror)


def make_readonly(path: PathType, stats: Optional[os.stat_result] = None) -> None:
    """deletes all write flags"""
    if not stats:
        stats = os.stat(path)
    os.chmod(path, stat.S_IMODE(stats.st_mode) & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)


def make_writeable(path: PathType, stats: Optional[os.stat_result] = None) -> None:
    """set owner write flag"""
    if not stats:
        stats = os.stat(path)
    os.chmod(path, stat.S_IMODE(stats.st_mode) | stat.S_IWRITE)


def is_writeable(stats: os.stat_result) -> bool:
    return stats.st_mode & stat.S_IWRITE != 0


def is_readable(stats: os.stat_result) -> bool:
    return stats.st_mode & stat.S_IREAD != 0


def isfile(stats: os.stat_result) -> bool:
    return stat.S_ISREG(stats.st_mode)


def isdir(stats: os.stat_result) -> bool:
    return stat.S_ISDIR(stats.st_mode)


# was: islink
def statsislink(stats: os.stat_result) -> bool:
    return stat.S_ISLNK(stats.st_mode)


def extract_basic_stat_info(stats: os.stat_result) -> Tuple[int, int, int, int]:
    """Returns size, modification, creation time and access mode."""
    return (stats.st_size, stats.st_mtime, stats.st_ctime, stats.st_mode)


def safe_filename(filename: str, replacement: str = "") -> str:
    WIN = windows_illegal_chars
    UNIX = unix_illegal_chars
    MAC = mac_illegal_chars
    bad = set.union(WIN, UNIX, MAC)

    safe_filename_translation_table = str.maketrans({c: replacement for c in bad})

    return filename.translate(safe_filename_translation_table)  # fixme: return callable which accepts only filename


safe_filename_simple = safe_filename


def _char_subber(s: str, illegal_chars: Set[str], replacement: str) -> str:
    if set(replacement) & illegal_chars:
        raise ValueError("replace character cannot be a illegal character")

    regex = "[" + re.escape("".join(illegal_chars)) + "]"
    return re.sub(regex, replacement, s)


def _char_splitter(s: str, chars: Iterable[str]) -> List[str]:
    regex = "[" + re.escape("".join(chars)) + "]"
    return re.split(regex, s)


def _special_subber(s: str, replacement: str = "_") -> str:
    if s == "." or s == "..":
        return s.replace(".", replacement)

    return s


def windows_compliant_filename(filename: str, replacement: str = "_") -> str:
    """https://msdn.microsoft.com/en-us/library/aa365247.aspx"""
    # reg. wikipedia: \x7f is valid!
    # trailing dots and spaces are ignored by windows

    ret = _char_subber(filename, windows_illegal_chars, replacement).rstrip(". ")  # strip or replace?, translate

    root, _ = os.path.splitext(ret)

    if root.upper() in windows_reserved_names:
        raise WindowsIllegalFilename("filename would result in reserved name")

    return ret


def windows_compliant_dirname(filename: str, replacement: str) -> str:
    """https://msdn.microsoft.com/en-us/library/aa365247.aspx"""
    # reg. wikipedia: \x7f is valid!
    # trailing dots and spaces are ignored by windows

    ret = _char_subber(filename, windows_illegal_chars, replacement).rstrip(" ")

    root, _ = os.path.splitext(ret)

    if root.upper() in windows_reserved_names:
        raise WindowsIllegalFilename("filename would result in reserved name")

    return ret


def linux_compliant_filename(filename: str, replacement: str = "_") -> str:
    filename = _special_subber(filename, replacement)
    return _char_subber(filename, unix_illegal_chars, replacement)


def linux_compliant_dirname(filename: str, replacement: str = "_") -> str:
    return _char_subber(filename, unix_illegal_chars, replacement)


def mac_compliant_filename(filename: str, replacement: str = "_") -> str:
    filename = _special_subber(filename, replacement)
    return _char_subber(filename, mac_illegal_chars, replacement)


def mac_compliant_dirname(filename: str, replacement: str = "_") -> str:
    return _char_subber(filename, mac_illegal_chars, replacement)


_os_map = {
    "compliant_filename": {
        "Windows": windows_compliant_filename,
        "Linux": linux_compliant_filename,
        "Darwin": mac_compliant_filename,
    },
    "compliant_dirname": {
        "Windows": windows_compliant_dirname,
        "Linux": linux_compliant_dirname,
        "Darwin": mac_compliant_dirname,
    },
}


def _get_os_func(name: str, force_system: Optional[str] = None) -> Callable:
    system = force_system or platform.system()

    return _os_map[name].get(system, _not_available(name))


compliant_filename = _get_os_func("compliant_filename")
compliant_dirname = _get_os_func("compliant_dirname")


def compliant_path(path: PurePath, replacement: str = "_", force_system: Optional[str] = None) -> PurePath:
    if force_system is None:
        _compliant_filename = compliant_filename
        _compliant_dirname = compliant_dirname
    else:
        _compliant_filename = _get_os_func("compliant_filename", force_system)
        _compliant_dirname = _get_os_func("compliant_dirname", force_system)

    fn_func = partial(_compliant_dirname, replacement=replacement)

    if bool(path.drive) or bool(path.root):  # is_absolute() doesn't do the right thing here
        parts = chain(path.parts[0:1], map(fn_func, path.parts[1:-1]))
    else:
        parts = map(fn_func, path.parts[:-1])

    filename = _compliant_filename(path.parts[-1], replacement)
    return type(path)(*parts, filename)


def reldirname(path: PathType) -> str:
    ret = os.path.dirname(path)
    if not ret:
        ret = "."

    return ret


def equal_dirs_iter(dir1: PathType, dir2: PathType, follow_symlinks: bool = False) -> Iterator[Tuple[str, str, str]]:
    """Tests if two directories are equal. Doesn't handle links."""

    def entry_path_size(entry) -> Tuple[str, int]:
        return entry.path, entry.stat().st_size

    files1 = sorted(map(entry_path_size, scandir_rec(dir1, follow_symlinks=follow_symlinks)))  # get rel paths
    files2 = sorted(map(entry_path_size, scandir_rec(dir2, follow_symlinks=follow_symlinks)))  # get rel paths

    for (path1, size1), (path2, size2) in zip_longest(files1, files2):
        if path1 != path2:
            yield ("name", path1, path2)

        elif size1 != size2:
            yield ("size", path1, path2)

    for (path1, size1), (path2, size2) in zip_longest(files1, files2):
        if path1 == path2 and size1 == size2 and not equal_files(path1, path2):
            yield ("data", path1, path2)


def equal_dirs(dir1: PathType, dir2: PathType) -> bool:
    return is_empty(equal_dirs_iter(dir1, dir2))


def search(
    directories: Iterable[PathType],
    pattern: str,
    dirs: bool = True,
    files: bool = True,
    rec: bool = True,
    follow_symlinks: bool = False,
) -> Iterator[MyDirEntryT]:
    """Search for files and folders matching the wildcard `pattern`."""

    for directory in directories:
        for entry in scandir_rec(directory, dirs=dirs, files=files, rec=rec, follow_symlinks=follow_symlinks):
            if fnmatch(entry.name, pattern):
                yield entry


def rename_files_in_folder(
    path: PathType, pattern: str, transform: Callable, template: str, rec: bool = True, follow_symlinks: bool = False
) -> None:
    cpattern = re.compile(pattern)

    for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
        filepath = entry.path
        base, name = os.path.split(filepath)

        m = cpattern.match(name)
        if not m:
            continue
        args = transform(*m.groups())

        to = template.format(*args)
        os.rename(filepath, os.path.join(base, to))


def clean_directory_by_extension(path: PathType, ext: str, rec: bool = True, follow_symlinks: bool = False) -> None:
    """Deletes all files of type `ext` if the same file without the ext exists
    and rename if it does not exit."""

    # delete .ext if normal file exists
    for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
        filepath = entry.path
        if not filepath.endswith(ext):
            file_to_delete = filepath + ext
            if os.path.isfile(file_to_delete):
                logger.info("Remove: %s", file_to_delete)
                try:
                    os.remove(file_to_delete)
                except OSError:
                    logger.exception("Could not delete: %s", file_to_delete)
                    break

    # rename all .ext to normal
    for entry in scandir_rec(path, files=True, dirs=False, rec=rec, follow_symlinks=follow_symlinks):
        filepath = entry.path
        if entry.name.endswith(ext):
            target = filepath[: -len(ext)]
            try:
                logger.info("Rename: %s -> %s", filepath, target)
                os.rename(filepath, target)
            except OSError:
                logger.exception("Should not have happened")
                break


def iter_links(path: PathType) -> Iterator[str]:
    """Yields all directory symlinks (and junctions on windows)."""

    for dirpath, dirnames, _filenames in os.walk(path):
        for dirname in dirnames:
            joined = os.path.join(dirpath, dirname)
            if islink(joined):
                yield joined


def normalize_seps(path: str) -> str:
    """Converts \\ to /"""

    return path.replace("\\", "/")


class Counts:
    __slots__ = ("dirs", "files", "others")

    def __init__(self) -> None:
        self.dirs = 0
        self.files = 0
        self.others = 0

    def __iadd__(self, other: "Counts") -> Self:
        self.dirs += other.dirs
        self.files += other.files
        self.others += other.others

        return self

    def null(self) -> bool:
        return self.dirs == 0 and self.files == 0 and self.others == 0


def _scandir_counts(
    rootentry: MyDirEntryT,
    files: bool = True,
    others: bool = True,
    rec: bool = True,
    total: bool = False,
    follow_symlinks: bool = True,
    errorfunc: Callable = scandir_error_raise,
) -> Iterator[Tuple[os.DirEntry, Optional[Counts]]]:
    counts = Counts()

    try:
        with os.scandir(rootentry.path) as it:
            for entry in it:
                if entry.is_dir(follow_symlinks=follow_symlinks):
                    counts.dirs += 1
                    if rec:
                        for subentry, subcounts in _scandir_counts(
                            entry, files, others, rec, total, follow_symlinks, errorfunc
                        ):
                            yield subentry, subcounts

                            if total:
                                assert subcounts
                                counts += subcounts

                elif entry.is_file(follow_symlinks=follow_symlinks):
                    counts.files += 1
                    if files:
                        yield entry, None

                else:
                    counts.others += 1
                    if others:
                        yield entry, None

        yield rootentry, counts  # yield after the loop

    except OSError as e:
        errorfunc(rootentry, e)


def scandir_counts(
    path: PathType,
    files: bool = True,
    others: bool = True,
    rec: bool = True,
    total: bool = False,
    follow_symlinks: bool = True,
    onerror: Callable = scandir_error_log,
) -> Iterator[Tuple[os.DirEntry, Optional[Counts]]]:
    """A recursive variant of scandir() which also returns the number of files/directories
    within directories.
    If total is True, the numbers will be calculated recursively as well.
    """

    if isinstance(path, os.PathLike):
        path = os.fspath(path)

    entry = DirEntryStub(os.path.basename(path), long_path_support(path))
    return _scandir_counts(entry, files, others, rec, total, follow_symlinks, onerror)


def shutil_onerror_remove_readonly(func, path, exc_info):
    """onerror function for the `shutil.rmtree` function.
    Makes read-only files writable and tries again.
    """

    stats = os.stat(path)
    if is_readable(stats):
        make_writeable(path, stats)
        func(path)
    else:
        raise


def _rmtree(
    path: str,
    ignore_errors: bool = False,
    onerror: Optional[Callable[[Callable, str, Any], None]] = None,
    follow_junctions: bool = False,
) -> Iterator[None]:
    """Compared to `shutil.rmtree`, this removes files instantly,
    instead of collecting the whole directory contents first.
    """

    import sys
    from shutil import _rmtree_isdir

    if not follow_junctions and sys.version_info < (3, 8):
        raise ValueError("`follow_junctions=False` is unsupported for Python<3.8.")

    if follow_junctions and sys.version_info >= (3, 8):
        raise ValueError("`follow_junctions=True` is unsupported for Python>=3.8.")

    if ignore_errors:

        def onerror(*args):
            pass

    elif onerror is None:

        def onerror(*args):
            raise

    try:
        with os.scandir(path) as scandir_it:
            for entry in scandir_it:
                fullname = entry.path
                if _rmtree_isdir(entry):
                    try:
                        if entry.is_symlink():  # not sure why this is necessary, copied from shutil._rmtree_unsafe
                            raise OSError("Cannot call _rmtree on a symbolic link")
                    except OSError:
                        onerror(os.path.islink, fullname, sys.exc_info())
                        continue
                    yield from _rmtree(fullname, onerror)
                else:
                    try:
                        os.unlink(fullname)
                        yield None
                    except OSError:
                        onerror(os.unlink, fullname, sys.exc_info())

    except OSError:
        onerror(os.scandir, path, sys.exc_info())

    try:
        os.rmdir(path)
        yield None
    except OSError:
        onerror(os.rmdir, path, sys.exc_info())



from copy import deepcopy
from typing import Collection, Generic, Iterator, List, Optional, Set, Tuple, TypeVar

from .compat.math import isqrt
from .exceptions import assert_type
from .indexing import col_indices, row_indices, subblock_indices
from .iter import batch
from .set import get as setget

T = TypeVar("T")


class Unsolvable(Exception):
    pass


class Sudoku(Generic[T]):
    def __init__(self, board: Collection[T], sym_set: Set[T], sym_free: T) -> None:
        assert_type("sym_set", sym_set, set)

        self.square = board
        self.sym_set = sym_set
        self.sym_free = sym_free

        self.outer_square_size = isqrt(len(board))
        self.outer_square_area = self.outer_square_size**2
        if self.outer_square_area != len(board):
            raise ValueError("board has an invalid size")

        if self.outer_square_size != len(sym_set):
            raise ValueError(
                f"sym_set length must be equal to the edge length of the board: {len(sym_set)} vs {self.outer_square_area}"
            )

        self.inner_square_size = isqrt(self.outer_square_size)
        if self.inner_square_size**2 != self.outer_square_size:
            raise ValueError("board has an invalid size")

        self.solved = False

    def init_board(self, board: Collection[T]) -> Collection[T]:
        raise NotImplementedError

    def get_board(self) -> Collection[T]:
        raise NotImplementedError

    def print_square(self) -> None:
        for i, num in enumerate(self.get_board(), 1):
            print(num, end=" ")
            if i % self.outer_square_size == 0:
                print()

    def solve(self, strategy: Optional[str] = None) -> Collection[T]:
        self.square = self.init_board(self.square)
        return self._solve(strategy)

    def _solve(self, strategy: Optional[str] = None) -> Collection[T]:
        raise NotImplementedError


class SudokuRulebased(Sudoku):
    def init_board(self, board: Collection[T]) -> List[Set[T]]:
        square = []
        for i in board:
            if i != self.sym_free:
                square.append({i})
            else:
                square.append(self.sym_set.copy())
        return square

    def get_board(self) -> Iterator[Set[T]]:
        for num in self.square:
            if len(num) == 1:
                yield setget(num)
            else:
                yield num

    def get_row_nums(self, i):
        row_set = set()
        for x in row_indices(i, self.outer_square_size):
            if len(self.square[x]) == 1 and x != i:
                row_set.update(self.square[x])
        return row_set

    def get_column_nums(self, i):
        column_set = set()
        for x in col_indices(i, self.outer_square_size):
            if len(self.square[x]) == 1 and x != i:
                column_set.update(self.square[x])
        return column_set

    def get_inner_square_nums(self, n):
        inner_square_set = set()
        for k in subblock_indices(n, self.outer_square_size, self.inner_square_size):
            if len(self.square[k]) == 1 and k != n:
                inner_square_set.update(self.square[k])
        return inner_square_set

    def update_cell(self, i):
        a = self.get_row_nums(i)
        b = self.get_column_nums(i)
        c = self.get_inner_square_nums(i)
        self.square[i] -= a
        self.square[i] -= b
        self.square[i] -= c

    def _solve(self, strategy=None):
        while not self.solved:
            self.solved = True
            square_old = deepcopy(self.square)
            for i in range(self.outer_square_area):
                if len(self.square[i]) != 1:
                    self.solved = False
                    # print_confirm("#%u is unsolved -> solve" % i)
                    self.update_cell(i)
                    # print_confirm(self.square)
            if square_old == self.square and not self.solved:
                raise NotImplementedError("Sudoku not solvable using this method")
        return self.square


class SudokuBruteforce(Sudoku):
    def init_board(self, board):
        return board

    def get_board(self):
        return self.square

    def get_row_nums(self, i):
        row_set = set()
        for x in row_indices(i, self.outer_square_size):
            if x != i:
                row_set.add(self.square[x])
        return row_set

    def get_column_nums(self, i):
        column_set = set()
        for x in col_indices(i, self.outer_square_size):
            if x != i:
                column_set.add(self.square[x])
        return column_set

    def get_inner_square_nums(self, i):
        inner_square_set = set()
        for x in subblock_indices(i, self.outer_square_size, self.inner_square_size):
            if x != i:
                inner_square_set.add(self.square[x])
        return inner_square_set

    def get_possible_nums(self, i: int) -> Set[T]:
        all = self.get_row_nums(i) | self.get_column_nums(i) | self.get_inner_square_nums(i)
        return self.sym_set - all

    def get_next_higher_possible_num(self, i: T, num: T) -> T:
        for pn in sorted(
            self.get_possible_nums(i)
        ):  # fixme: sorting is bad here. # why? for algorithmic performance or python performance?
            if pn > num:
                return pn

        return self.sym_free

    def get_next_lower_possible_num(self, i: T, num: T) -> T:
        for pn in sorted(self.get_possible_nums(i), reverse=False):  # see: get_next_higher_possible_num
            if pn < num:
                return pn

        return self.sym_free

    def _solve(self, strategy: Optional[str] = None) -> Tuple[int, int]:
        strategy = strategy or "inc"

        try:
            candidate_func = {
                "inc": self.get_next_higher_possible_num,
                "dec": self.get_next_lower_possible_num,
            }[strategy]

            selection_func = {
                "inc": min,
                "dec": max,
            }[strategy]

        except KeyError:
            raise ValueError("Invalid strategy")

        i = 0
        backtrack: Tuple[int, T] = []
        backtracks = 0
        steps = 0

        while i < self.outer_square_area:
            if self.square[i] != self.sym_free:
                i += 1
                continue
            pnums = self.get_possible_nums(i)

            if len(pnums) == 0:
                while True:
                    backtracks += 1
                    try:
                        j, num = backtrack.pop()
                    except IndexError:
                        raise Unsolvable

                    num = candidate_func(j, num)
                    self.square[j] = num
                    if num != self.sym_free:
                        backtrack.append((j, num))
                        i = j + 1
                        break
            else:
                num = selection_func(pnums)
                self.square[i] = num
                backtrack.append((i, num))
                steps += 1

        return backtracks, steps


def is_valid_solution(board: Collection[T], sym_set: Set[T]) -> bool:
    """Checks if `board` is a valid solved Sudoku configuration"""

    edge_len = isqrt(len(board))
    square_len = isqrt(edge_len)

    board = tuple(batch(board, edge_len, func=tuple))

    # check rows
    for j in range(edge_len):
        row = {board[j][i] for i in range(edge_len)}
        if set(row) != sym_set:
            return False

    # check cols
    for i in range(edge_len):
        col = {board[j][i] for j in range(edge_len)}
        if col != sym_set:
            return False

    # check inner squares
    for x in range(0, edge_len, square_len):
        for y in range(0, edge_len, square_len):
            square = {board[i][j] for i in range(x, x + square_len) for j in range(y, y + square_len)}
            if square != sym_set:
                return False

    return True


if __name__ == "__main__":
    from argparse import ArgumentParser

    from genutility.time import MeasureTime

    parser = ArgumentParser()
    parser.add_argument("board", metavar="N", type=int, nargs="+", help="Flat Sudoku board")
    parser.add_argument("--symbols", default=set(range(1, 10)), type=int, nargs="+", help="Used symbols")
    parser.add_argument("--free", default=0, help="Free marker symbol")
    parser.add_argument("--strategy", choices=("inc", "dec"), default="linear")
    args = parser.parse_args()

    sym_set = set(args.symbols)

    s = SudokuBruteforce(args.board, sym_set, args.free)

    try:
        with MeasureTime() as t:
            steps, backtracks = s.solve(args.strategy)

        s.print_square()
        if is_valid_solution(args.board, sym_set):
            print(f"Solved sudoku in {t.get():.2} seconds, using {steps} steps and backtracked {backtracks} times.")
        else:
            print(
                f"No valid solution was found in {t.get():.2} seconds, using {steps} steps and backtracked {backtracks} times."
            )
    except Unsolvable:
        print("Sudoku is not solvable")



from functools import wraps
from typing import Callable
from warnings import warn


def deprecated(msg: str, stacklevel: int = 2) -> Callable[[Callable], Callable]:
    def decorator(func):
        @wraps(func)
        def inner(*args, **kwargs):
            warn(msg, DeprecationWarning, stacklevel)  # noqa: B028
            print("DeprecationWarning:", msg)
            return func(*args, **kwargs)

        return inner

    return decorator



from itertools import islice
from math import pi

import cv2
import numpy as np
from sklearn.cluster import MeanShift

from .math import inf
from .rand import rgb_colors

## UNFINISHED
## BASED ON: Resizing by Symmetry-Summarization (2010)


def MSER_draw(img, boxes, centers, major_v, minor_v, labels):
    center_int = centers.astype(int)
    major_int = (centers + major_v).astype(int)
    minor_int = (centers + minor_v).astype(int)

    labels_unique = np.unique(labels)
    n_clusters = len(labels_unique)
    colors = list(islice(rgb_colors(), n_clusters))
    print(f"{n_clusters} clusters found")

    for box, label in zip(boxes, labels):
        cv2.ellipse(img, box, colors[label], 2)

    for c, ma, mi in zip(center_int, major_int, minor_int):
        cv2.line(img, tuple(c), tuple(ma), (0, 255, 0), 2)
        cv2.line(img, tuple(c), tuple(mi), (0, 255, 0), 2)


def MSER_boxes(img):
    mser = cv2.MSER_create()

    msers, _ = mser.detectRegions(img)

    return [cv2.fitEllipse(mser) for mser in msers]


def MSER_ellipses(boxes):
    pihalf = pi / 2.0
    boxes = np.array([(x, y, ma, mi, a) for (x, y), (ma, mi), a in boxes])

    centers = boxes[:, 0:2]
    angles = boxes[:, 4] * 2.0 * pi / 360.0  # convert degrees to rad

    major_x = boxes[:, 2] / 2 * np.cos(angles)
    major_y = boxes[:, 2] / 2 * np.sin(angles)
    major_v = np.stack([major_x, major_y], axis=-1)

    angles += pihalf
    minor_x = boxes[:, 3] / 2 * np.cos(angles)
    minor_y = boxes[:, 3] / 2 * np.sin(angles)
    minor_v = np.stack([minor_x, minor_y], axis=-1)

    return (centers, major_v, minor_v)


def MSER(img, show=True):
    boxes = MSER_boxes(img)
    centers, major_v, minor_v = MSER_ellipses(boxes)

    return boxes, centers, major_v, minor_v


def norm(v):
    return np.sqrt(np.sum(v**2, axis=-1))


def shape_dissimmilarity(R_u, R_v, i, j):
    num1 = np.abs(R_u[i] - R_u[j])
    denom1 = max(norm(R_u[i]), norm(R_u[j]))

    num2 = np.abs(R_v[i] - R_v[j])
    denom2 = max(norm(R_v[i]), norm(R_v[j]))

    return num1 / denom1 + num2 / denom2


def appearance_dissimmilarity():
    pass


def symmetry_detection(image, K_c=5):
    # Identifying cells with MSER
    boxes, R_c, R_u, R_v = MSER(image)  # {r_i} set of MSERs (2d ellipsis) r_i = (center, major axis, minor axis)

    # Clustering with adaptive mean-shift clustering
    u_norm = norm(R_u)
    v_norm = norm(R_v)
    shapessizes = np.stack([u_norm, v_norm], axis=-1)

    ms = MeanShift()
    labels = ms.fit_predict(shapessizes)

    MSER_draw(image, boxes, R_c, R_u, R_v, labels)
    cv2.imshow("img", image)
    cv2.waitKey()


def diff(v1, v2, tau_v):
    val = abs(v1 - v2) / max(norm(v1), norm(v2))
    if val > tau_v:
        return inf
    else:
        return val


if __name__ == "__main__":
    imgpath = "D:\\colloseo.jpg"
    img = cv2.imread(imgpath)
    symmetry_detection(img)



import platform
from typing import NamedTuple

_usagetuple = NamedTuple("usage", [("total", int), ("used", int), ("free", int)])
_volumeinfotuple = NamedTuple(
    "volumeinfo",
    [
        ("VolumeName", str),
        ("VolumeSerialNumber", int),
        ("MaximumComponentLength", int),
        ("FileSystemFlags", int),
        ("FileSystemName", str),
    ],
)


def is_os_64bit():
    return platform.machine().endswith("64")



import dbm
import gzip
import logging
import pickle  # nosec
import warnings
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Callable, ContextManager, Iterator, Optional, Tuple

import scrapy

from .dbm import dbm_items
from .stdio import print_terminal_progress_line

if TYPE_CHECKING:
    from collections.abc import MutableMapping


def read_dbm_httpcache(
    path: str, open_func: Callable[[str, str], ContextManager[MutableMapping]] = dbm.open, decompress: bool = True
) -> Iterator[Tuple[bytes, float, Any]]:
    """Loads scrapy dbm http cache files.
    Uses pickle so only use on trusted file.
    """

    if decompress:
        import brotli

    with open_func(path, "r") as db:
        for key, value in dbm_items(db):
            if key.endswith(b"_data"):
                hash = key[:-5]
                time = float(db[hash + b"_time"])
                data = pickle.loads(value)  # nosec

                if decompress:
                    ce = data["headers"].get(b"Content-Encoding", [])
                    if ce == [b"gzip"]:
                        data["body"] = gzip.decompress(data["body"])
                        data["headers"][b"Content-Encoding"] = []
                    elif ce == [b"br"]:
                        data["body"] = brotli.decompress(data["body"])
                        data["headers"][b"Content-Encoding"] = []
                    elif ce:
                        warnings.warn(f"Unsupported Content-Encoding: {ce}", stacklevel=2)

                yield hash, time, data


def print_progress(spider: scrapy.Spider) -> None:
    queue = len(spider.crawler.engine.slot.scheduler)
    requests = len(spider.crawler.engine.slot.scheduler.df.fingerprints)
    delta = datetime.utcnow() - spider.crawler.stats.get_value("start_time")
    items = spider.crawler.stats.get_value("item_scraped_count", 0)
    files = spider.crawler.stats.get_value("file_status_count/downloaded", 0)
    warnings = spider.crawler.stats.get_value("log_count/ERROR", 0)
    errors = spider.crawler.stats.get_value("log_count/WARNING", 0)
    cachehit = spider.crawler.stats.get_value("httpcache/hit", 0)
    cachemiss = spider.crawler.stats.get_value("httpcache/miss", 0)

    deltaseconds = timedelta(seconds=int(delta.total_seconds()))

    out = (
        f"{deltaseconds} requests: {requests:07d}, queue: {queue:07d}, items: {items:07d}, files: {files:07d}"
        f", warn/err: {warnings:05d}/{errors:05d}, cache: {cachehit:07d}/{cachehit+cachemiss:07d}"
    )
    print_terminal_progress_line(out)


def get_url_logger(urllogfile: Optional[str], name: str = "urllog", funcname: bool = False) -> logging.Logger:
    if urllogfile:
        handler = logging.FileHandler(urllogfile, encoding="utf-8", delay=True)
    else:
        handler = logging.NullHandler()

    urllog = logging.getLogger(name)
    urllog.propagate = False
    urllog.setLevel(logging.INFO)

    handler.setLevel(logging.INFO)
    if funcname:
        fmtstring = "%(asctime)s\t%(funcName)10s\t%(message)s"
    else:
        fmtstring = "%(asctime)s\t%(message)s"
    formatter = logging.Formatter(fmtstring)
    handler.setFormatter(formatter)
    urllog.addHandler(handler)
    return urllog


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("path", help="path to db file")
    args = parser.parse_args()

    for _hash, _time, data in read_dbm_httpcache(args.path):
        print(data["url"][-50:], len(data["body"]), data["body"][:30])



from collections import Counter
from itertools import chain
from math import exp, log2
from typing import Any, Collection, Dict, Iterable, Iterator, List, MutableMapping, Optional, Tuple, Union

import numpy as np

from .callbacks import Progress
from .datastructures.sparse_matrix import VariableRowMatrix
from .encoder import BatchLabelEncoder
from .numpy import batchtopk, categorical
from .object import cast

RawDocument = List[int]
IterableDocuments = Iterable[Collection[int]]
TopicsMapping = MutableMapping[Tuple[int, int], int]
Indices = Union[slice, np.ndarray]

_SLICE_NONE = slice(None)


def top_topics(
    id2word: Dict[int, str], term_topic_matrix: np.ndarray, num_words: int = 10, decimals: int = 2
) -> Iterator[List[Tuple[str, float]]]:
    """
    term_topic_matrix: float[K, V]
    """

    indices_list, probs_list = batchtopk(term_topic_matrix, num_words, reverse=True)
    np.around(probs_list, decimals=decimals, out=probs_list)

    for indices, probs in zip(indices_list, probs_list):
        yield [(id2word[id], prob) for id, prob in zip(indices, probs)]


def format_topics(topics: Iterable[List[Tuple[str, float]]], linesep: str = "\n", tokensep: str = "\t") -> str:
    buffer: List[str] = []

    for terms in topics:
        buffer.append(tokensep.join(f"({term},{prob})" for term, prob in terms))

    return linesep.join(buffer)


class LDADocument(Collection[int]):
    def __init__(self, words: RawDocument) -> None:
        self.words = words

    def __iter__(self) -> Iterator[int]:
        return iter(self.words)

    def __len__(self) -> int:
        return len(self.words)

    def __repr__(self) -> str:
        return f"LDADocument({repr(self.words)})"


class LDABase:
    K: int
    nmk: np.ndarray
    nkt: np.ndarray
    nk: np.ndarray
    nm: np.ndarray

    α: float
    β: float
    αsum: float
    βsum: float

    docs: List[LDADocument]

    def __init__(self, seed: Optional[int] = None) -> None:
        np.seterr(all="raise")
        np.random.seed(seed)

    def init_topic(self) -> int:
        return np.random.randint(self.K)

    def add_counts(self, m: int, t: int, k: int) -> None:
        self.nmk[m, k] += 1
        self.nkt[k, t] += 1
        self.nk[k] += 1

    def sub_counts(self, m: int, t: int, k: int) -> None:
        self.nmk[m, k] -= 1
        self.nkt[k, t] -= 1
        self.nk[k] -= 1

    def _validate(self, M: int, K: int, V: int) -> None:
        if M <= 0:
            raise ValueError("Number of documents must be larger than zero")
        if K <= 0:
            raise ValueError("Number of topics must be larger than zero")
        if V <= 0:
            raise ValueError("Vocabulary size must be larger than zero")

        assert self.nmk.shape == (M, K)
        assert self.nm.shape == (M,)
        assert self.nkt.shape == (K, V)
        assert self.nk.shape == (K,)

    def _initialize_topics(self, docs: IterableDocuments, topics: TopicsMapping, progress: Progress) -> None:
        for m, doc in enumerate(progress.track(docs)):
            for i, t in enumerate(doc):
                k = topics[m, i] = self.init_topic()
                self.add_counts(m, t, k)

    def calc_probs(self, m: int, t: int, k_select: Indices = _SLICE_NONE) -> np.ndarray:
        # notes: np.sum(self.nkt, axis=-1) == self.nk
        # notes: np.sum(self.nmk[m], axis=-1) == self.nm[m]

        """This is eq1. in the DC-LDA paper.

        returns: float[K]
        """

        """ The full term for left would be:
            left = (self.nmk[m, :] + self.α) / (self.nm[m] + self.αsum) # [K]
            however, the denominator can be dropped, because it doesn't depend on `k`
            and we only care about proportionality.
            It is done so in: 'A Theoretical and Practical Implementation Tutorial
                on Topic Modeling and Gibbs Sampling'
        """

        left = self.nmk[m, k_select] + self.α  # [K]
        right = (self.nkt[k_select, t] + self.β) / (self.nk[k_select] + self.βsum)  # [K]
        return left * right  # [K]

    def sample(self, m: int, t: int) -> int:
        probs = self.calc_probs(m, t)  # [K]
        return categorical(probs)

    def _sample_all(self, docs: IterableDocuments, topics: TopicsMapping, progress: Progress) -> None:
        for m, doc in enumerate(progress.track(docs)):
            for i, t in enumerate(doc):
                k = topics[m, i]
                self.sub_counts(m, t, k)
                k = topics[m, i] = self.sample(m, t)  # sample new topics
                self.add_counts(m, t, k)

    def initialize(self) -> None:
        """Set base parameters"""

        raise NotImplementedError

    def initialize_docs(self) -> None:
        """Set parameters based on input docs"""

        raise NotImplementedError

    def initialize_topics(self) -> None:  # step1
        """Initialize topics matrix"""

        raise NotImplementedError

    def sample_all(self) -> None:  # step2 and step3
        """Sample new topics"""

        raise NotImplementedError

    def metric(self) -> float:
        raise NotImplementedError

    def fit(self, n_iter: int = 100, verbose: bool = False) -> None:
        """Fit LDA"""

        assert self.docs, "No documents added to model"

        self.initialize()
        self.initialize_docs()
        self.initialize_topics()

        for i in range(n_iter):
            self.sample_all()
            if verbose:
                print(f"Step #{i}, metric: {self.metric()}")

    def theta(self) -> np.ndarray:
        """
        returns: float[M, K]
        """

        num = self.nmk + self.α  # [M, K]
        denom = self.nm + self.αsum  # [M]
        return num / denom[:, None]  # [M, K]

    def phi(self) -> np.ndarray:
        """
        returns: float[K, V]
        """

        num = self.nkt + self.β  # [K, V]
        denom = self.nk + self.βsum  # [K]
        return num / denom[:, None]  # [K, V]

    def _perplexity(self, docs: IterableDocuments) -> float:
        θ = self.theta()
        φ = self.phi()

        num = 0
        denom = 0

        for m, doc in enumerate(docs):
            for i, t in enumerate(doc):
                num += np.log(np.inner(φ[:, t], θ[m]))
            denom += len(doc)

        return exp(-num / denom)

    def docs2topics(self) -> np.ndarray:
        """
        returns: int[M]
        """

        return np.argmax(self.theta(), axis=-1)  # [M]

    def print_topics(self, id2word: Dict[int, str], num_words: int = 10) -> None:
        topics = top_topics(id2word, self.phi(), num_words)
        print(format_topics(topics))

    @staticmethod
    def generate(theta: np.ndarray, phi: np.ndarray, doc_lens: Iterable[int]) -> Tuple[dict, dict]:
        """
        theta: float[K, V]
        phi: float[M, K]
        """

        z = {}  # sparse
        w = {}  # sparse

        for m, length in enumerate(doc_lens):
            for i in range(length):
                k = z[(m, i)] = categorical(theta[m])  # scalar # doc to topic
                w[(m, i)] = categorical(phi[k])  # scalar # topic to word

        return z, w

    def inferencer(self) -> "LDAInfer":
        return cast(self, LDAInfer)


class LDA(LDABase):
    """LDA using collapsed Gibbs sampling
    Based on:
    - https://en.wikipedia.org/wiki/Latent_Dirichlet_allocation
    - Latent Dirichlet Allocation (2003)
    - A Theoretical and Practical Implementation Tutorial on Topic Modeling and Gibbs Sampling (2011)
    - Integrating Out Multinomial Parameters in Latent Dirichlet Allocation and Naive Bayes for Collapsed Gibbs Sampling (2010)
    """

    def __init__(
        self,
        n_topics: int,
        alpha: float = 0.1,
        beta: float = 0.01,
        seed: Optional[int] = None,
        progress: Optional[Progress] = None,
    ) -> None:
        LDABase.__init__(self, seed)

        self.K = n_topics  # number of topics

        self.alpha = alpha
        self.beta = beta

        self.word_encoder: BatchLabelEncoder[str] = BatchLabelEncoder(tokenizer="nltk")
        self.docs: List[LDADocument] = []

        self.inttype = np.int32
        self.progress = progress or Progress()

    def initialize(self) -> None:
        self.V = self.word_encoder.num_labels

        self.α = self.alpha / self.K  # symmetric Dirichlet prior for document-topic distribution θ
        self.β = self.beta / self.V  # symmetric Dirichlet prior for topic-word distribution φ

        self.αsum = self.alpha
        self.βsum = self.beta

        self.nkt = np.zeros((self.K, self.V), dtype=self.inttype)  # [K, V] topic-word counts
        self.nk = np.zeros((self.K,), dtype=self.inttype)  # [K] total topic counts

    def initialize_docs(self) -> None:
        self.M = len(self.docs)

        self.nmk = np.zeros((self.M, self.K), dtype=self.inttype)  # [M, K] document-topic counts
        self.nm = np.array(
            list(map(len, self.docs)), dtype=self.inttype
        )  # [M] total document counts, not necessary for sampling, but useful for theta and phi calculation

        # note: VariableRowMatrix(0) is slower than a dict here for some reason
        self.topics: TopicsMapping = (
            {}
        )  # `z`, sparse document-word topics matrix (because rows can have different lengths)

    def initialize_topics(self) -> None:
        self._validate(self.M, self.K, self.V)

        return self._initialize_topics(self.docs, self.topics, self.progress)

    def sample_all(self) -> None:
        return self._sample_all(self.docs, self.topics, self.progress)

    def metric(self) -> float:
        return self._perplexity(self.docs)

    def perplexity(self, docs: IterableDocuments) -> float:
        return self._perplexity(docs)

    def transform(self, doc_lens: Iterable[int]) -> Tuple[dict, dict]:
        return self.generate(self.theta(), self.phi(), doc_lens)

    def add_doc(self, doc: str) -> int:
        assert isinstance(doc, str)

        words = list(self.word_encoder.fit_transform(doc))

        ret = len(self.docs)
        self.docs.append(LDADocument(words))
        return ret

    def make_doc(self, doc: str) -> LDADocument:
        assert isinstance(doc, str)

        words = list(self.word_encoder.transform(doc))
        return LDADocument(words)


class LDATermWeight(LDABase):
    """LDA with term weighting
    Based on: "Term Weighting Schemes for Latent Dirichlet Allocation" (2010)

    Further analysis papers:
            - "Pulling Out the Stops: Rethinking Stopword Removal for Topic Models" (2017)
            - "Assessing topic model relevance: Evaluation and informative priors" (2019)
    """

    def __init__(
        self,
        n_topics: int,
        tws: str = "PMI",
        alpha: float = 0.1,
        beta: float = 0.01,
        progress: Optional[Progress] = None,
    ) -> None:
        """tws: Term weighting scheme
        ONE: Consider every term equal (same as standard LDA)
        TF: Term frequency
        IDF: Inverse Document Frequency term weighting.
                Thus, a term occurring at almost every document has very low weighting and a term occurring at a few document has high weighting.
        PMI: Pointwise Mutual Information term weighting (default)
        """

        LDABase.__init__(self)

        self.K = n_topics  # number of topics

        self.alpha = alpha
        self.beta = beta

        self.word_encoder: BatchLabelEncoder[str] = BatchLabelEncoder(tokenizer="nltk")
        self.docs: List[LDADocument] = []

        self.create_term_weights = {
            "ONE": self._one,
            "TF": self._tf,
            "IDF": self._idf,
            "TFIDF": self._tfidf,
            "PMI": self._pmi,
            "RAND": self._rand,
        }[tws]

        self.floattype = np.float32
        self.inttype = np.int32

        self.progress = progress or Progress()

    def _one(self, docs: Any) -> np.ndarray:
        """
        returns: int[M, V]
        """

        # return np.ones((self.M, self.V))
        return np.broadcast_to(1, (self.M, self.V))  # is this slower for indexing later?

    def _tf(self, docs: IterableDocuments) -> np.ndarray:
        """
        returns: float[M, V]
        """

        all_counts = Counter(chain.from_iterable(docs))
        num_counts = sum(all_counts.values())

        tw = np.array([all_counts[t] for t in range(self.V)], dtype=self.floattype)  # [V]
        tw = -np.log2(tw / num_counts)

        return np.broadcast_to(tw, (self.M, self.V))

    def _idf(self, docs: IterableDocuments) -> np.ndarray:
        """
        returns: float[M, V]
        """

        tw = np.zeros((self.V,), dtype=self.floattype)  # [V]

        for doc in docs:
            for token in set(doc):
                tw[token] += 1.0

        tw = np.log2(self.M / tw)

        return np.broadcast_to(tw, (self.M, self.V))

    def _tfidf(self, docs):
        raise NotImplementedError

    def _pmi(self, docs: IterableDocuments) -> np.ndarray:
        """
        returns: float[M, V]
        """

        tw = np.zeros((self.M, self.V), dtype=self.floattype)  # [M, V] term weights

        all_counts = Counter(chain.from_iterable(docs))

        for m, doc in enumerate(docs):
            doc_counts = Counter(doc)
            for i, t in enumerate(doc):
                tw[m, t] = -log2(doc_counts[t] / all_counts[t])

        np.clip(tw, 0.0, None, out=tw)
        return tw

    def _rand(self, docs: Any) -> np.ndarray:
        """
        returns: float[M, V]
        """

        return np.random.uniform(0, 1, (self.M, self.V)).astype(self.floattype)

    def initialize(self) -> None:
        self.V = self.word_encoder.num_labels

        self.α = self.alpha / self.K  # symmetric Dirichlet prior for document-topic distribution θ
        self.β = self.beta / self.V  # [V] symmetric Dirichlet prior for topic-word distribution φ

        self.αsum = self.alpha
        self.βsum = self.beta

        self.nkt = np.zeros((self.K, self.V), dtype=self.inttype)  # [K, V] counts, same as np.sum(nmkt, axis=0)
        self.nk = np.zeros((self.K,), dtype=self.inttype)  # [K] counts, same as np.sum(nkt, axis=-1) # [K]

        # self.wkt = np.zeros((self.K, self.V), dtype=self.floattype) # [K, T] weights
        # self.wk = np.zeros((self.K, ), dtype=self.floattype) # [K] weights

    def initialize_docs(self) -> None:
        self.M = len(self.docs)

        self.tw = self.create_term_weights(self.docs)
        print(self.tw.shape)
        print(self.tw)

        self.nmk = np.zeros((self.M, self.K), dtype=self.inttype)  # [M, K] counts, same as np.sum(nmkt, axis=-1)
        self.nm = np.array(
            list(map(len, self.docs)), dtype=self.inttype
        )  # [M] total document counts, not necessary for sampling, but useful for theta and phi calculation

        """
            `wmk[m]` is same as `np.sum(wmkt[m], axis=-1)` is same as
            `np.sum(self.tw[m, :][None, :] * nmkt[m], axis=-1)`

            with
                `wmkt = np.zeros((self.M, self.K, self.V), dtype=self.floattype)` # [M, K, V] weights
                `nmkt = np.zeros((self.M, self.K, self.V), dtype=self.inttype)` # [M, K, V] counts
        """
        # self.nmkt = np.zeros((self.M, self.K, self.V), dtype=self.inttype)
        self.wmk = np.zeros((self.M, self.K), dtype=self.floattype)  # [M, K] weights, same as np.sum(nmkt, axis=-1)

        self.topics: VariableRowMatrix[int] = VariableRowMatrix(
            0
        )  # [M, Variable] `z`, sparse document-word topics matrix

    def initialize_topics(self) -> None:
        # self._validate(self.M, self.K, self.V)

        self._initialize_topics(self.docs, self.topics, self.progress)

    def calc_probs(self, m: int, t: int) -> np.ndarray:
        """
        returns: float[K]
        """

        # here using global weights works
        # wmk = np.sum(self.tw[m, :][None, :] * self.nmkt[m], axis=-1, dtype=self.floattype)
        # if not np.allclose(self.wmk[m], wmk, atol=1e-06): # after some time it still diverges because of nummerical issues
        #     raise RuntimeError("inconsistent wmk")

        # here it doesn't...
        wkt = self.tw[m, t] * self.nkt[:, t]
        # if not np.allclose(self.wkt[:, t], wkt[:]):
        #     raise RuntimeError("inconsistent wkt")

        # here it doesn't also...
        wk = np.sum(self.tw[m, :][None, :] * self.nkt, axis=-1)  # not the same as `self.wk[k] (+|-)= self.tw[m, t]`
        # if not np.allclose(self.wk, wk):
        #     raise RuntimeError("inconsistent wk")

        # nk: ld.numByTopic[tid]
        # nkt: ld.numByTopicWord[tid, vid]

        left = self.wmk[m] + self.α  # [K]
        right = (wkt[:] + self.β) / (wk[:] + self.βsum)  # [K]
        return left * right  # [K]

    """
    def addWordToINC(docId, tid, vid):

        # paper W*beta term in eq. 6 (https://www.aclweb.org/anthology/N10-1070.pdf)
        ld.numByTopicWordCOUNTS[tid, vid] += 1
        Nk = np.sum(docs[docId].wordWeights[None,:] * ld.numByTopicWordCOUNTS, axis=-1) # [None,:] needed for correct broadcasting. summed over words, K size array remains
        # or the above in plain python
        Nk = [sum(docs[docId].wordWeights[word_id] * ld.numByTopicWordCOUNTS[tid, word_id]
            for word_id in docs[docid].words)
                for tid in range(K)] # K size array

        # addWordTo INC
        ld.numByTopic[tid] += docs[docId].wordWeights[pid] # ld.numByTopic is a K size array

        # now both weight sums should be the same AT ALL TIMES
        assert Nk == ld.numByTopic # elementwise comparison

        # but they are not after the first document, if the weights depend on the document
    """

    def add_counts(self, m: int, t: int, k: int) -> None:
        self.nmk[m, k] += 1
        self.nkt[k, t] += 1
        self.nk[k] += 1

        # self.nmkt[m, k, t] += 1
        self.wmk[m, k] += self.tw[m, t]
        # self.wkt[k, t] += self.tw[m, t]
        # self.wk[k] += self.tw[m, t]

        """
        print("docId={}, vid={}, tid={}".format(m, t, k))

        wk = np.sum(self.tw[m, :][None, :] * self.nkt, axis=-1)
        print("ld.numByTopicWordCOUNTS:\n", self.nkt)
        print("paper sum:", wk)
        print("ld.numByTopic:", self.wk)

        if not np.allclose(wk, self.wk):
            raise RuntimeError("not equal")
        """

    def sub_counts(self, m: int, t: int, k: int) -> None:
        self.nmk[m, k] -= 1
        self.nkt[k, t] -= 1
        self.nk[k] -= 1

        # self.nmkt[m, k, t] -= 1
        self.wmk[m, k] -= self.tw[m, t]
        # self.wkt[k, t] -= self.tw[m, t]
        # self.wk[k] -= self.tw[m, t]

    def sample_all(self) -> None:
        self._sample_all(self.docs, self.topics, self.progress)

    def metric(self) -> float:
        return self._perplexity(self.docs)


class LDAInfer(LDA):
    def __init__(self) -> None:
        self.nkt_old = self.nkt
        self.nk_old = self.nk
        self.docs = []

    def calc_probs(self, m: int, t: int, k_select: Indices = _SLICE_NONE) -> np.ndarray:
        """
        returns: float[K]
        """

        left = self.nmk[m, k_select] + self.α  # [K]
        num = self.nkt[k_select, t] + self.nkt_old[k_select, t] + self.β  # [K]
        denom = self.nk[k_select] + self.nk_old[k_select] + self.βsum  # [K]

        return left * num / denom  # [K]

    def initialize(self) -> None:
        assert self.K > 0, "Number of topics must be larger than zero"
        assert self.V > 0, "Vocabulary size must be larger than zero"

        self.nkt = np.zeros((self.K, self.V), dtype=self.inttype)
        self.nk = np.zeros((self.K,), dtype=self.inttype)


def test_lda_20():
    from sklearn.datasets import fetch_20newsgroups

    ng = fetch_20newsgroups()

    lda = LDA(n_topics=20, seed=0)

    for doc in ng.data[:1000]:
        lda.add_doc(doc)

    lda.fit(10, verbose=True)

    return lda


def test_ldatw_20():
    from sklearn.datasets import fetch_20newsgroups

    ng = fetch_20newsgroups()

    lda = LDATermWeight(n_topics=20)

    for doc in ng.data[:1000]:
        lda.add_doc(doc)

    lda.fit(10, verbose=True)

    return lda


if __name__ == "__main__":
    from genutility.time import PrintStatementTime

    for i in range(3):
        with PrintStatementTime():
            test_lda_20()
            # test_ldatw_20()



from typing import Any, Callable, Container, Sequence, TypeVar

T = TypeVar("T")


def extract_indexes(d: dict) -> Callable[[Sequence], dict]:
    return lambda seq: {name: seq[index] for index, name in d.items()}


def bit_or(x: Any, y: Any) -> Any:
    return x | y


def not_in(s: Container[T]) -> Callable[[T], bool]:
    return lambda x: x not in s


def operator_in(s: Container[T]) -> Callable[[T], bool]:
    """Comparable to stdlib `operator.itemgetter` type functions.
    It returns a function which will return a boolean indicating if
    its argument if in `s`. `s` should be a set-like for better performance.
    """

    return lambda x: x in s


def logical_xor(a: Any, b: Any) -> bool:
    # this is just the != operator for bools

    return not (a and b) and (a or b)


def logical_xnor(a: Any, b: Any) -> bool:
    # this is just the == operator for bools

    return not (a or b) or (a and b)


def logical_implication(a: Any, b: Any) -> bool:
    """if a is True, b must be True"""

    return (not a) or b


def logical_nonimplication(a: Any, b: Any) -> bool:
    """if a is True, b must be False"""

    return a and (not b)


def converse_implication(a: Any, b: Any) -> bool:
    """if a is False, b must be False"""

    return a or (not b)


def converse_nonimplication(a: Any, b: Any) -> bool:
    """if a is False, b must be True"""

    return (not a) and b



import tensorflow as tf


def categorical(probs, size=1):
    logits = tf.math.log(probs)[None, :]
    return tf.squeeze(tf.random.categorical(logits, num_samples=size), axis=0)


def randint(high, size=(), dtype=tf.int64):
    return tf.random.uniform(size, maxval=high, dtype=dtype)



from typing import Any

from ruamel.yaml import YAML

from ._files import PathType


def read_yaml(path: PathType) -> Any:
    with open(path, encoding="utf-8") as fr:
        yaml = YAML(typ="safe")
        return yaml.load(fr)


def write_yaml(obj: PathType, path: Any) -> None:
    with open(path, "w", encoding="utf-8") as fw:
        yaml = YAML(typ="safe")
        yaml.dump(obj, fw)
